import argparse

class _genebed12(object):
    """bed12 object"""
    
    def __init__(self,chrom,start,end,name,strand, gene_biotype):
        self.chrom = chrom
        self.start = start
        self.end = end
        self.name = name
        self.score = 0
        self.strand = strand
        self.thickStart = end
        self.thickEnd = end
        self.itemRgb = 0
        self.blockCount = 0
        self.exon_site = {}
        self.gene_biotype = gene_biotype
        #self.blockSizes = {}
        #self.blockStarts = {}

    def add_exon_block(self,s,e):        
        #self.blockCount += 1
        self.exon_site["{0}-{1}".format(s,e)] = 0
        
            
    def tostr(self, biotype2bed13):
        blockSizes = None
        blockStarts = None
        blockCount = len(self.exon_site)
        
        for i,v in self.exon_site.items():
            tmp = i.split("-")
            if blockSizes == None:
                blockSizes = str(int(tmp[1]) - int(tmp[0]))
                blockStarts = tmp[0]
            else:
                blockSizes = blockSizes + "," + "{0}".format(int(tmp[1]) - int(tmp[0]))
                blockStarts = blockStarts + "," + tmp[0]
        
      
        obj_str = "{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\t{7}\t{8}\t{9}\t{10}\t{11}\t{12}".format(self.chrom, self.start, self.end,\
                                                                                        self.name, self.score, self.strand,\
                                                                                        self.thickStart, self.thickEnd, self.itemRgb,\
                                                                                blockCount, blockSizes, blockStarts, self.gene_biotype)
        if biotype2bed13 == False:
             obj_str = "{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\t{7}\t{8}\t{9}\t{10}\t{11}".format(self.chrom, self.start, self.end,\
                                                                                        self.name, self.score, self.strand,\
                                                                                        self.thickStart, self.thickEnd, self.itemRgb,\
                                                                                blockCount, blockSizes, blockStarts)
        return(obj_str)
            

def _bedcol9_2dic(bedcol9, split=";"):
    tmp = bedcol9.split(split)
    col9dic = {}
    for i in tmp:
        i = i.strip()
        ii = i.split(" ")
        val = ii[1].strip("\"")
        col9dic[ii[0]] = val
    return col9dic



gene = None
#gene_id="gene_id"
gene_name_tag="gene_name"
gene_id_tag="gene_id"
gene_biotype_tag="gene_biotype"
split = ";"
biotype2bed13 = True


def _parse_arguments(args=None):
    """
    get command line arguments
    """
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=False,
        description='Converts the GTF to bed12 or bed13')

    parserRequired = parser.add_argument_group('Required arguments')

    # define the arguments
    parserRequired.add_argument('--gtf', '-g',
                                help='input file. GTF format gene annotation.',
                                required=True)

    parserRequired.add_argument('--outFileName', '-o',
                                help='File name of gene annotation bed12 format.',
                                required=True)

    parserOpt = parser.add_argument_group('Optional arguments')
    parserOpt.add_argument('--splitby', '-s',
                           help='tags of GTF split by which str',
                           default=";")
    parserOpt.add_argument('--gene_name_tag', '-nt',
                           help='tags of GTF gene_name',
                           default="gene_name")
    parserOpt.add_argument('--gene_id_tag', '-it',
                           help='tags of GTF gene_id',
                           default="gene_id")
    parserOpt.add_argument('--gene_biotype_tag', '-bt',
                           help='tags of GTF gene_biotype',
                           default="gene_biotype")
    parserOpt.add_argument('--biotype2bed13', '-bed13',
                           help='if this arg is set, the out file columns 13 will add gene\'s biotype',
                           action='store_true')

    return parser


def _main(args=None):

    args = _parse_arguments().parse_args(args)
    gene_name_tag = args.gene_name_tag
    gene_id_tag = args.gene_id_tag
    gene_biotype_tag = args.gene_biotype_tag
    split = args.splitby
    biotype2bed13 = args.outFileName

    gene = None
    gtf_hand = open(args.gtf, "r")
    gtf2bed = open(args.outFileName,'w')

    for line in gtf_hand:
        bedtab = line.rstrip().split('\t')
        if len(bedtab) < 9:
            continue

        #if re.match('chr', bedtab[0]) == None:
        #    continue
            
        #if re.match('chrM', bedtab[0]) != None:
        #    continue
            
        #print(bedtab)
        #print(gene)
            
        if bedtab[2] == "gene":
            if gene != None:
                gtf2bed.write(gene.tostr(biotype2bed13) + "\n")
            Gene_name = 'xx'
            Gene_biotype = "yy"
            
            
            col9dic = _bedcol9_2dic(bedtab[8], split=split)
            if gene_name_tag in col9dic.keys():
                Gene_name = col9dic[gene_name_tag]
            elif gene_id_tag in col9dic.keys():
                Gene_name = col9dic[gene_id_tag]
            else:
                print(bedtab[8], ' have no {0}, xx instead'.format(gene_name_tag))
            
            if gene_biotype_tag in col9dic.keys():
                Gene_biotype = col9dic[gene_biotype_tag]
            else:
                print(bedtab[8], ' have no {0}, yy instead'.format(gene_biotype_tag))
                
            gene = _genebed12(bedtab[0], bedtab[3], bedtab[4], Gene_name, bedtab[6], Gene_biotype)
        
        elif bedtab[2] == "exon":
            if gene != None:
                gene.add_exon_block(bedtab[3],bedtab[4])
        else:
            pass
            
    #最后一行没有输出
            
    gtf_hand.close()
    gtf2bed.close()
    print ('gtf2bed finished')



from matplotlib.axes import Axes
from typing import Union, Optional, Sequence
import pandas as pd
from trackc.tl import GenomeRegion
from trackc.pa import trackcl_11

def _scale_ticks(ticks, scale='bp', tick_fl='%0.2f'):
    if scale == 'Mb':
        ticks = ticks/1000000
        ticks = ["{0}".format(tick_fl % i) for i in ticks]
    elif scale == 'kb': 
        ticks = ticks/1000
        ticks = ["{0}".format(tick_fl % i) for i in ticks]
    else:
        return ticks
    return ticks
def scale_track(ax: Optional[Axes] = None,
                region: Union[str, None] = None,
                tick_pos: str = 'bottom',
                ratio2ax: float = 0.5,
                chrom_fontsize: Union[int, None] = 10,
                scale_adjust: Union[str, None] = 'kb',
                tick_fl: Union[str, None] ='%0.2f',
                tick_fontsize: Union[int, None] = 8,
                tick_rotation: Union[int, None] = 0,
                space: float = 0.1,
                ):
    """
    Parameters
    ----------
    
    scale_adjust
        ``str``: should be one of [kb, Mb, bp]
            adjust the scale unit to make it pretty
    tick_fl
        ``str``: The position label retains the number of decimal places  
    tick_pos
        one of ['top', 'bottom', 'left', 'right']                   

    """
    line_GenomeRegions = None
    if isinstance(region, list):
        print('scale_track is only for one region')
        return
    else:
        line_GenomeRegions = GenomeRegion(region).GenomeRegion2df()

    line_GenomeRegions = line_GenomeRegions.reset_index()

    chrom = line_GenomeRegions.loc[0, "chrom"]
    start = line_GenomeRegions.loc[0, "start"]
    end = line_GenomeRegions.loc[0, "end"]

    pos_dic = {
        'left': [-ratio2ax,	0, ratio2ax, 1],
        'right': [1, 0,	ratio2ax, 1],
        'top': [0, 1+space, 1, ratio2ax],
        'bottom': [0, -ratio2ax-space, 1, ratio2ax]
    }

    ax2 = ax.inset_axes(pos_dic[tick_pos], facecolor='none')
    ax = ax2
    
    if tick_pos in ['top', 'bottom']:
        ax.set_xlim([start, end])

        xticks = ax.get_xticks()
        xtick_labels = xticks
        if scale_adjust == 'Mb':
            xtick_labels = xtick_labels/1000000
            xtick_labels = ["{0}".format(tick_fl % i) for i in xtick_labels]
            ax.set_xticks(xticks, xtick_labels, fontsize=tick_fontsize, rotation=tick_rotation)
            ax.spines['bottom'].set_position(('data', 0))
            #ax.text(end-(abs(end-start)*0.05), -1, 'Mb', fontsize=chrom_fontsize)
        elif scale_adjust == 'kb':
            xtick_labels = xtick_labels/1000
            xtick_labels = ["{0}".format(tick_fl % i) for i in xtick_labels]
            ax.set_xticks(xticks, xtick_labels, fontsize=tick_fontsize, rotation=tick_rotation)
            #ax.text(end-(abs(end-start)*0.05), -1, 'kb', fontsize=chrom_fontsize)
            ax.spines['bottom'].set_position(('data', 0))
        else:
            pass

        spines = ['bottom', 'top', 'right', 'left']
        chrom_x = start + (end-start)/2
        chrom_y = 1
        va = 'top'
        
        if tick_pos=='bottom':
            del spines[1]
            ax.tick_params(top=True, labeltop=True, bottom=False, labelbottom=False)
            chrom_y = 0
            va = 'bottom'
            
        if tick_pos=='top':
            del spines[0]
            ax.tick_params(top=False, labeltop=False, bottom=True, labelbottom=True)
        for i in spines:
            ax.spines[i].set_visible(False)

        ax.text(chrom_x, chrom_y, chrom, fontsize=chrom_fontsize, ha='center', va=va)
        labels = [label.get_text() for label in ax.get_xticklabels()]
        labels[-1] += '({0})'.format(scale_adjust)
        ax.set_xticklabels(labels)

        ax.tick_params(which='major', direction='in', pad=-16) 
        ax.set_yticks([])
        ax.set_yticklabels('')  
    


def multi_scale_track(ax: Optional[Axes] = None,
                      regions: Union[Sequence[str], str, None] = None,
                      colors: Union[Sequence[str], None] = None,
                      alpha: Union[float, None] = 1,
                      intervals: Union[int, None] = 1,
                      scale_adjust: Union[str, None] = 'kb',
                      tick_fl: Union[str, None] ='%0.2f',
                      tick_fontsize: Union[int, None] = 8,
                      tick_rotation: Union[int, None] = 0,
                ):
    """\
    """

    if colors != None:
        track_colors  = colors
    else:
        track_colors = trackcl_11

    if isinstance(regions, list):
        line_GenomeRegions = pd.concat([GenomeRegion(i).GenomeRegion2df() for i in regions])
    else:
        line_GenomeRegions = GenomeRegion(regions).GenomeRegion2df()

    line_GenomeRegions = line_GenomeRegions.reset_index()

    if len(track_colors) < line_GenomeRegions.shape[0]:
        repeat_times = (line_GenomeRegions.shape[0] + len(track_colors) - 1) // len(track_colors)
        track_colors = (track_colors * repeat_times)[:line_GenomeRegions.shape[0]]
    else:
        track_colors = track_colors[:line_GenomeRegions.shape[0]]
    line_GenomeRegions['color'] = track_colors
    line_GenomeRegions['len'] = line_GenomeRegions['fetch_end'] - line_GenomeRegions['fetch_start']
    line_GenomeRegions['acum'] = line_GenomeRegions['len'].cumsum()
    sum_len = line_GenomeRegions['len'].sum()
    bottom = list(range(intervals))  
    # broadcast
    if intervals < line_GenomeRegions.shape[0]:
        repeat_times = (line_GenomeRegions.shape[0] + intervals - 1) // intervals
        bottom = (bottom * repeat_times)[:line_GenomeRegions.shape[0]]
    else:
        bottom = bottom[:line_GenomeRegions.shape[0]]
    line_GenomeRegions.loc[:, 'bottom'] = bottom

    line_GenomeRegions['tick_s'] = _scale_ticks(line_GenomeRegions['start'], scale=scale_adjust, tick_fl=tick_fl)
    line_GenomeRegions['tick_e'] = _scale_ticks(line_GenomeRegions['end'], scale=scale_adjust, tick_fl=tick_fl)
    if scale_adjust in ['Mb', 'kb']:
        line_GenomeRegions.loc[line_GenomeRegions.index[-1], "tick_e"] = line_GenomeRegions.loc[line_GenomeRegions.index[-1], "tick_e"] + "(" + scale_adjust + ")"

    for i, row in line_GenomeRegions.iterrows():
        arrow_s = row['acum'] - row['len']
        dx = row['len']
        if row["isReverse"] == True:
            arrow_s = row['acum']
            dx = -1 * row['len']
            
        ax.arrow(arrow_s, row['bottom'], dx, 0, 
            overhang=1, width=0.01,
            head_width=0.25,
            head_length=sum_len/170,
            length_includes_head=True,
            color=row['color'],
            linewidth=1,
            alpha = alpha
            )
        
        tick_ha = 'center'
        if tick_rotation != 0:
            tick_ha = 'right'

        ax.text(row['acum'] - row['len']/2, row['bottom']+0.1, row['chrom'], ha='center', va='bottom', fontsize=tick_fontsize, color=row['color'])
        ax.text(row['acum'] - row['len'], row['bottom']-0.1, row['tick_s'], ha=tick_ha, va='top', fontsize=tick_fontsize, rotation=tick_rotation, color=row['color'])
        ax.text(row['acum'], row['bottom']-0.1, row['tick_e'], ha=tick_ha, va='top', fontsize=tick_fontsize, rotation=tick_rotation, color=row['color'])


    ax.set_xlim([0, sum_len])
    ax.set_ylim([-0.7, intervals-0.3])

    for i in ['bottom', 'top','right', 'left']:
        ax.spines[i].set_color('none')
        ax.spines[i].set_linewidth(0)

    ax.tick_params(bottom =False,top=False,left=False,right=False)
    ax.set_xticklabels("")
    ax.set_yticklabels("")
    #ax.spines["bottom"].set_color('black')
    #ax.spines["bottom"].set_linewidth(1)













#
# Sphinx-Gallery documentation build configuration file, created by
# sphinx-quickstart on Mon Nov 17 16:01:26 2014.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys
import os
from datetime import date
import warnings

#import sphinx_gallery
from sphinx_gallery.sorting import FileNameSortKey
from sphinx_gallery.notebook import add_markdown_cell, add_code_cell

HERE = os.path.abspath(os.path.dirname(__file__))
package_dir = os.path.join(HERE, '..', 'src')
sys.path.insert(0, package_dir)

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#sys.path.insert(0, os.path.abspath('.'))

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.mathjax',
    'sphinx.ext.todo',
    #'sphinx_gallery',
    'sphinx.ext.coverage',
    'sphinx_gallery.load_style',
    'sphinx.ext.graphviz',
    'nbsphinx',
    'IPython.sphinxext.ipython_console_highlighting',
    'sphinxcontrib.bibtex',
    'sphinx.ext.autosummary',
    'sphinx_copybutton',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# generate autosummary even if no references


# The suffix of source filenames.
source_suffix = [".rst", ".md"]

# The encoding of source files.
#source_encoding = 'utf-8-sig'

copyright = '%s, Zan Yuan' % date.today().year

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = '0.1.21'
# The full version, including alpha/beta/rc tags.
release = version + '-git'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
#today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build', '**.ipynb_checkpoints']

# See warnings about bad links
nitpicky = True
nitpick_ignore = [('', "Pygments lexer name 'ipython' is not known")]

# The reST default role (used for this markup: `text`) to use for all
# documents.
#default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
#add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#show_authors = False

# The name of the Pygments (syntax highlighting) style to use.


# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
#keep_warnings = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

# The theme is set by the make target

project = 'TrackC'
master_doc = 'index'
html_theme = 'pydata_sphinx_theme'
autosummary_generate = True
html_show_sphinx =  False
pygments_style = 'sphinx'
highlight_language = 'python3'



# bibliography
bibtex_bibfiles = ["references.bib"]
bibtex_reference_style = "author_year"
bibtex_default_style = "alpha"

nbsphinx_thumbnails = {
    'thumbnail_size': '400x300',
    'examples/zoomin_heatmap': 'examples/zoomin_heatmap.png',
    'tutorials/bw_track_tutorial': 'examples/zoomin_heatmap.png',
}

def setup(app):
    """Sphinx setup function."""
    app.add_css_file('theme_override.css')
    app.add_object_type('confval', 'confval',
                        objname='configuration value',
                        indextemplate='pair: %s; configuration value')

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.


html_theme_options = {
    'navbar_center': ['navbar-nav'],
    'show_toc_level': 2,
    'show_nav_level': 2,
    #'navbar_end': ['theme-switcher', 'version-switcher', 'navbar-icon-links'],
    'logo': {
        'text': 'TrackC',
    },
    #'switcher': dict(
    #    json_url='https://sphinx-gallery.github.io/dev/_static/switcher.json',
    #    version_match='dev' if 'dev' in version else 'stable',
    #),
    "github_url": "https://github.com/seqyuan/trackc",
    "icon_links": [
        {
            "name": "PyPI",
            "url": "https://pypi.org/project/trackc",
            "icon": "fa-solid fa-box",
        },
    ],
}

# Add any paths that contain custom themes here, relative to this directory.
#html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
#html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
#html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
#html_favicon = None


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
#html_extra_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
#html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
html_sidebars = {
    'reference': [],
}

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
#html_domain_indices = True

# If false, no index is generated.
#html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
#html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
#html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = 'Trackc-Gallerydoc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #'preamble': '',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    ('index', 'Sphinx-Gallery.tex', 'Sphinx-Gallery Documentation',
     'Óscar Nájera', 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
#latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
#latex_use_parts = False

# If true, show page references after internal links.
#latex_show_pagerefs = False

# If true, show URL addresses after external links.
#latex_show_urls = False

# Documents to append as an appendix to all manuals.
#latex_appendices = []

# If false, no module index is generated.
#latex_domain_indices = True


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('index', 'sphinx-gallery', 'Sphinx-Gallery Documentation',
     ['Óscar Nájera'], 1)
]

# If true, show URL addresses after external links.
#man_show_urls = False


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    ('index', 'Sphinx-Gallery', 'Sphinx-Gallery Documentation',
     'Óscar Nájera', 'Sphinx-Gallery', 'One line description of project.',
     'Miscellaneous'),
]

# Documents to append as an appendix to all manuals.
#texinfo_appendices = []

# If false, no module index is generated.
#texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
#texinfo_show_urls = 'footnote'

# If true, do not generate a @detailmenu in the "Top" node's menu.
#texinfo_no_detailmenu = False

"""
nbsphinx_custom_formats = {
    '.pct.py': ['jupytext.reads', {'fmt': 'py:percent'}],
    '.md': ['jupytext.reads', {'fmt': 'Rmd'}],
}
"""
# Example configuration for intersphinx: refer to the Python standard library.


examples_dirs = ['../tutorials']
gallery_dirs = ['../examples']


image_scrapers = ('matplotlib',)

min_reported_time = 0
if 'SOURCE_DATE_EPOCH' in os.environ:
    min_reported_time = sys.maxint if sys.version_info[0] == 2 else sys.maxsize

"""
def notebook_modification_function(notebook_content, notebook_filename):
    notebook_content_str = str(notebook_content)
    warning_template = "\n".join(
        [
            "<div class='alert alert-{message_class}'>",
            "",
            "# JupyterLite warning",
            "",
            "{message}",
            "</div>"
        ]
    )

    if "pyvista_examples" in notebook_filename:
        message_class = "danger"
        message = (
            "PyVista is not packaged in Pyodide, this notebook is not "
            "expected to work inside JupyterLite"
        )
    elif "import plotly" in notebook_content_str:
        message_class = "danger"
        message = (
            "This notebook is not expected to work inside JupyterLite for now."
            " There seems to be some issues with Plotly, see "
            "[this]('https://github.com/jupyterlite/jupyterlite/pull/950') "
            "for more details."
        )
    else:
        message_class = "warning"
        message = (
            "JupyterLite integration in sphinx-gallery is beta "
            "and it may break in unexpected ways"
        )

    markdown = warning_template.format(
        message_class=message_class,
        message=message
    )

    dummy_notebook_content = {'cells': []}
    add_markdown_cell(dummy_notebook_content, markdown)

    code_lines = []

    if "seaborn" in notebook_content_str:
        code_lines.append("%pip install seaborn")

    if code_lines:
        code_lines = ["# JupyterLite-specific code"] + code_lines
        code = "\n".join(code_lines)
        add_code_cell(dummy_notebook_content, code)

    notebook_content["cells"] = (
        dummy_notebook_content["cells"] + notebook_content["cells"]
    )


from pathlib import Path
_root = Path(__file__).parent.parent
sphinx_gallery_conf = {
    'examples_dirs': [_root / "tutorials"],
    'gallery_dirs': [_root / "examples"],
}



sphinx_gallery_conf = {
    'backreferences_dir': 'gen_modules/backreferences',
    'doc_module': ('sphinx_gallery', 'numpy'),
    'reference_url': {
        'sphinx_gallery': None,
    },
    'examples_dirs': examples_dirs,
    'gallery_dirs': gallery_dirs,
    'image_scrapers': image_scrapers,
    'compress_images': ('images', 'thumbnails'),
    # specify the order of examples to be according to filename
    'within_subsection_order': FileNameSortKey,
    #'expected_failing_examples': ['../examples/no_output/plot_raise.py',
    #                              '../examples/no_output/plot_syntaxerror.py'],
    'min_reported_time': min_reported_time,
    'binder': {'org': 'sphinx-gallery',
               'repo': 'sphinx-gallery.github.io',
               'branch': 'master',
               'binderhub_url': 'https://mybinder.org',
               'dependencies': './binder/requirements.txt',
               'notebooks_dir': 'notebooks',
               'use_jupyter_lab': True,
               },
    #'jupyterlite': {
    #    'notebook_modification_function': notebook_modification_function
    #},
    'show_memory': True,
    'promote_jupyter_magic': False,
    'junit': os.path.join('sphinx-gallery', 'junit-results.xml'),
    # capture raw HTML or, if not present, __repr__ of last expression in
    # each code block
    'capture_repr': ('_repr_html_', '__repr__'),
    'matplotlib_animations': True,
    'image_srcset': ["2x"],
    'nested_sections': True,
    'show_api_usage': True,
}
"""



# Remove matplotlib agg warnings from generated doc when using plt.show
warnings.filterwarnings("ignore", category=UserWarning,
                        message='Matplotlib is currently using agg, which is a'
                                ' non-GUI backend, so cannot show the figure.')




from matplotlib.axes import Axes
from typing import Union, Optional, Sequence
import pandas as pd
from trackc.tl._getRegionsCmat import GenomeRegion
from .bigwig import _make_multi_region_ax


def gene_track(ax: Optional[Axes] = None,
               gene_bed: pd.DataFrame = None,
               regions: Union[Sequence[str], str, None] = None,
               track_type: Union[str, None] = 'gene',
               pos_strand_gene_color: Union[str, None] = '#3366CC',
               neg_strand_gene_color: Union[str, None] = '#EECFA1',
               line: Union[int, None] = 1,
               gene_fontszie: Union[int, None] = 5,
               label: Optional[str] = None,
               label_rotation: Union[int, None] = 0,
               label_fontsize: Optional[int] = 12,
               ):
    """\
    Plot multi-regions gene track.
    
    Parameters
    ----------
    gene_bed
        ``pd.DataFrame``:
    ax

    regions

    track_type

    pos_strand_gene_color

    neg_strand_gene_color

    line

    gene_fontszie

    ylabel

    label_rotation

    label_fontsize

    """

    if isinstance(regions, list):
        line_GenomeRegions = pd.concat([GenomeRegion(i).GenomeRegion2df() for i in regions])
    else:
        line_GenomeRegions = GenomeRegion(regions).GenomeRegion2df()

    axs = _make_multi_region_ax(ax, line_GenomeRegions)
    line_GenomeRegions = line_GenomeRegions.reset_index()

    ax.set_ylabel(label, fontsize=label_fontsize, rotation=label_rotation, 
                  horizontalalignment='right',verticalalignment='center')
    spines = ['top', 'bottom', 'left', 'right']
    for i in spines:
        ax.spines[i].set_visible(False)
    ax.set_xticks([])
    ax.set_xticklabels('')
    ax.set_yticks([])
    ax.set_yticklabels('')  

    gene_bed.columns =['chrom','start','end','name',"score","strand","thickStart","thickEnd","itemRgb","blockCount","blockSizes","blockStarts"]
    for ix, row in line_GenomeRegions.iterrows(): 
        if track_type == "gene":
             _plot_gene(axs[ix], gene_bed, row['chrom'], row['fetch_start'], row['fetch_end'], needReverse=row['isReverse'], pos_strand_gene_color=pos_strand_gene_color, neg_strand_gene_color=neg_strand_gene_color, line=line, fontszie=gene_fontszie)
        if track_type == "density":
             print('This gene type is developping')
        else:
             pass
def _plot_gene(ax, gene_bed, chrom, start, end, needReverse=False, pos_strand_gene_color='#3366CC', neg_strand_gene_color='#EECFA1', line=1, fontszie=5):
    gene_bed = gene_bed[gene_bed['chrom']==chrom]
    gene_bed_plot = gene_bed[((gene_bed['start'] >= start) & (gene_bed['start'] <= end)) | ((gene_bed['end'] >= start) & (gene_bed['end'] <= end))]
    gene_bed_plot = gene_bed_plot.sort_values(by='end')
    #print(gene_bed_plot
    
    plot_gene_num = gene_bed_plot.shape[0]

    ii = 0
    for i,row in gene_bed_plot.iterrows():
        #col = pos_strand_gene_color
        text_col = pos_strand_gene_color
        
        if row["strand"] == "-":
            #col = neg_strand_gene_color
            text_col = neg_strand_gene_color
        
        #text_col = col
        plot_y = ii%line
        
        ax.plot((row['start'], row['end']), (plot_y + 0.5, plot_y+0.5), color='k', linewidth=1, solid_capstyle='butt')
        starts = [int(x) for x in row["blockStarts"].split(",")]
        widths = [int(x) for x in row["blockSizes"].split(",")]
        
        ax.bar(x=starts, height=0.4, width=widths, bottom=plot_y+0.3, \
                edgecolor='k', linewidth=1, align='edge', color='k')
        
        if row['start'] < start:
                row['start'] = start
        if row['end'] > end:
                row['end'] = end
        
        arrow_s = row['end']
        dx = 0.1
        if row["strand"] == "-":
            arrow_s = row['start']
            dx = -0.1
        ax.arrow(arrow_s, plot_y+0.5, dx, 0, 
            overhang=0, width=0,
            head_width=0.25,
            head_length=10000,
            length_includes_head=False,
            color=text_col,
            linewidth=1)

        if (row['name'] in gene_bed_plot.iloc[-4:,:]['name']) or (int(line/(ii+1)) < 2):
            ha = 'right'
            genename = row['name'] + "  "
            xpos = row['start']
            if needReverse:
                ha = 'left'
                genename = "  " + row['name']
                #xpos = row['end']
            ypos = plot_y + 0.5
            if line == 1:
                 xpos = row['start'] + abs(row['start']-row['end'])/2
                 ypos = 0.8
                 ha = 'center'
            ax.text(xpos, ypos, genename + "  ", ha=ha, va='center',color=text_col, fontsize=fontszie)
        else:
            ha = 'left'
            genename = "  " + row['name']
            xpos = row['end']
            if needReverse:
                ha = 'right'
                genename = row['name'] + "  "
                #xpos = row['start']

            ypos = plot_y + 0.5
            if line == 1:
                 xpos = row['start'] + abs(row['start']-row['end'])/2
                 ypos = 0.8
                 ha = 'center'
            ax.text(xpos, ypos, genename, ha=ha, va='center',color=text_col, fontsize=fontszie)

        ii+=1
            
    xlim_s = start
    xlim_e = end
    if needReverse == True:
         xlim_s = end
         xlim_e = start

    ax.set_xlim(xlim_s, xlim_e)
    ax.set_ylim(top=0, bottom=line)
    if plot_gene_num < line:
        ax.spines['bottom'].set_position(('data', plot_gene_num))
    
    for i in ['left','top','right']:
        ax.spines[i].set_color('none')
        ax.spines[i].set_linewidth(0)
    ax.spines["bottom"].set_color('black')
    ax.spines["bottom"].set_linewidth(0.5)
    ax.tick_params(bottom =True,top=False,left=False,right=False)
    ax.set_xticklabels("")
    ax.set_yticklabels("")




from matplotlib.axes import Axes
from typing import Union, Optional, Sequence
from matplotlib.patches import Polygon
from matplotlib.collections import PatchCollection
from matplotlib.colors import Colormap
import matplotlib.pyplot as plt
from matplotlib import cm
import pandas as pd
import numpy as np
from trackc.tl._getRegionsCmat import GenomeRegion
from .bigwig import _make_multi_region_ax

def bed_track(ax: Optional[Axes] = None,
              bed: pd.DataFrame = None,
              regions: Union[Sequence[str], str, None] = None,
              track_style: Union[str, None] = 'bar',
              color: Union[Sequence[str], None] = 'tab:blue',
              cmap: Union[Sequence[Colormap], str, None] = None,
              intervals: Union[int, None] = 1,
              #show_names: Union[bool, None] = False,
              alpha: Union[float, None] = 1,
              label: Union[str, None] = None,
              label_fontsize: Union[int, None] = 5,
              label_rotation: Union[int, None] = 0,
              ymin: Optional[float] = None,
              ymax: Optional[float] = None,
              tick_fontsize: Optional[int] = 8,
              tick_fl: Optional[str] ='%0.2f',
              score_label_size: Union[int, None] = 7,
              ):
    """
    Parameters
    ----------
    bed
        ``pd.DataFrame``: 6th columns dataframe
            should be sorted py chromStart if track_style is line
            column1: chrom
                The name of the chromosome (e.g. chr3, chrY, chr2_random) or scaffold (e.g. scaffold10671).
            column2: chromStart
                The starting position of the feature in the chromosome or scaffold. The first base in a chromosome is numbered 0.
            column3: chromEnd
                The ending position of the feature in the chromosome or scaffold.
            column4: name
                Defines the name of the BED line. Either "." (=no name), or other string
            column5: score
                Defines the name of the BED line. if nessasary, can be set as ``.``,
                if track_type/style is one of bar/line/heatmap 
            column6: strand 
                Defines the strand. Either "." (=no strand) or "+" or "-".
    track_style
        ``str``: bed blocks style, 
            line
            bar
            heatmap
            triangle
            rec
                rectangle
            arrow
            
    color
        ``str`` or list: the color of line/triangle/rectangle, if color is color list, the block will 
    intervals
        ``int``: if track_style is one of [triangle, rec], the row number distribution for triangle/rectangle blocks
    """
    if isinstance(regions, list):
        line_GenomeRegions = pd.concat([GenomeRegion(i).GenomeRegion2df() for i in regions])
    else:
        line_GenomeRegions = GenomeRegion(regions).GenomeRegion2df()

    axs = _make_multi_region_ax(ax, line_GenomeRegions)
    line_GenomeRegions = line_GenomeRegions.reset_index()
    if isinstance(color, list)==False:
        color = [color]
    if len(color) < line_GenomeRegions.shape[0]:
        repeat_times = (line_GenomeRegions.shape[0] + len(color) - 1) // len(color)
        color = (color * repeat_times)[:line_GenomeRegions.shape[0]]

    if isinstance(cmap, list)==False:
        cmap = [cmap]
    if len(cmap) < line_GenomeRegions.shape[0]:
        repeat_times = (line_GenomeRegions.shape[0] + len(cmap) - 1) // len(cmap)
        cmap = (cmap * repeat_times)[:line_GenomeRegions.shape[0]]

    ax.set_ylabel(label, fontsize=label_fontsize, rotation=label_rotation, ha='right', va='center')
    spines = ['top', 'right', 'left']
    for i in spines:
        ax.spines[i].set_visible(False)
    ax.set_xticks([])
    ax.set_xticklabels('')
    ax.set_yticks([])
    ax.set_yticklabels('')  

    score_label = None
    if bed.shape[1] == 3:
        bed.columns = ['chrom', 'start', 'end']
    if bed.shape[1] == 4:
        bed.columns = ['chrom', 'start', 'end', 'name']
    if bed.shape[1] == 5:
        score_label = bed.columns[4]
        bed.columns = ['chrom', 'start', 'end', 'name', 'score']
    if bed.shape[1] == 6:
        score_label = bed.columns[4]
        bed.columns = ['chrom', 'start', 'end', 'name', 'score', 'strand']

    min_y = None
    max_y = None
    max_len = 0
    for ix, row in line_GenomeRegions.iterrows(): 
        bed2plot = bed[(bed['chrom']==row['chrom']) & (bed['end']>=row['fetch_start']) & (bed['start']<=row['fetch_end'])]
        if track_style in ['line', 'bar'] or bed.shape[1]>=5:
            if min_y==None:
                min_y = bed2plot['score'].min()
            else:
                if min_y < bed2plot['score'].min():
                    min_y = bed2plot['score'].min()
        
            if max_y==None:
                max_y = bed2plot['score'].max()
            else:
                if max_y > bed2plot['score'].max():
                    max_y = bed2plot['score'].max()

        maxlength = (bed2plot['end']-bed2plot['start']).max()
        if max_len < maxlength:
            max_len = maxlength

    if ymin == None:
        ymin = min_y
    if ymax == None:
        ymax = max_y

    for ix, row in line_GenomeRegions.iterrows(): 
        bed2plot = bed[(bed['chrom']==row['chrom']) & (bed['end']>=row['fetch_start']) & (bed['start']<=row['fetch_end'])].copy()
        if track_style == "line":
            plot_bed_bar_l(axs[ix], bed2plot, row['fetch_start'], row['fetch_end'], needReverse=row['isReverse'], style='line', color=color[ix], alpha=alpha)    
        if track_style == "bar":
            _plot_bed_bar_l(axs[ix], bed2plot, row['fetch_start'], row['fetch_end'], needReverse=row['isReverse'], style='bar', color=color[ix], alpha=alpha)
        if track_style == "rec":
            _plot_bed_rec(ax, axs[ix], bed2plot, row['fetch_start'], row['fetch_end'], 
                         needReverse=row['isReverse'], color=color[ix], cname=cmap[ix], 
                         alpha=alpha, min=ymin, max=ymax, score_label=score_label, intervals=intervals, score_label_size=score_label_size)
        if track_style == "triangle":
            _plot_bed_tri(ax, axs[ix], bed2plot, row['fetch_start'], row['fetch_end'], 
                         needReverse=row['isReverse'], color=color[ix], cname=cmap[ix], 
                         alpha=alpha, min=ymin, max=ymax, score_label=score_label, score_label_size=score_label_size)
            
            axs[ix].set_ylim(0, max_len/2)

    if track_style in ['line', 'bar']:
        for axi in axs:
            axi.set_ylim(ymin, ymax)
        ax.set_ylim(ymin, ymax)
        ax.text(0, ymax, " [{0}, {1}]".format(tick_fl % ymin, tick_fl % ymax), va='top', fontsize=tick_fontsize)


def _make_tri_data(start, end):
    data = np.array([[start, 0], [end, 0], [start+(end-start)/2, (end-start)/2]])
    return data

def _plot_bed_tri(mainAX, ax, bed, start, end, needReverse, color, cname, alpha, min, max, score_label=None, score_label_size=8):
    colors = color
    norm = None
    if cname != None:
        if isinstance(cname, str):
            map_vir=cm.get_cmap(cname)
        if isinstance(cname, Colormap):
            map_vir=cname
        
        # 因为 y 大到一定程度超过临界数值后颜色就会饱和不变(不使用循环colormap)。
        norm = plt.Normalize(min, max)
        # matplotlib.colors.Normalize 对象，可以作为参数传入到绘图方法里
        # 也可给其传入数值直接计算归一化的结果
        norm_y = norm(bed['score'])
        colors = map_vir(norm_y)
        
    patches = []
    bed = bed.reset_index()
    #print(colors[0])
    #print(colors[1])

    for i, row in bed.iterrows():
        polygon = Polygon(_make_tri_data(row['start'], row['end']), True, color=colors)
        patches.append(polygon)
    
    p = PatchCollection(patches, alpha=alpha, match_original=True)
    ax.add_collection(p)
    
    xlim_s = start
    xlim_e = end
    if needReverse == True:
         xlim_s = end
         xlim_e = start

    ax.set_xlim(xlim_s, xlim_e)
    if cname != None:
        sm = cm.ScalarMappable(norm=norm, cmap=map_vir)
        cax = mainAX.inset_axes([1.01, 0, 0.01, 1])
        cb = plt.colorbar(sm, ax=mainAX, cax=cax, label=score_label)
        cb.set_label(score_label, fontsize=score_label_size)

def _plot_bed_rec(mainAX, ax, bed, start, end, needReverse, color, cname, alpha, min, max, score_label=None, intervals=1, score_label_size=8):
    colors = color
    if cname != None:
        if isinstance(cname, str):
            map_vir=cm.get_cmap(cname)
        if isinstance(cname, Colormap):
            map_vir=cname
        
        # 因为 y 大到一定程度超过临界数值后颜色就会饱和不变(不使用循环colormap)。
        norm = plt.Normalize(min, max)
        # matplotlib.colors.Normalize 对象，可以作为参数传入到绘图方法里
        # 也可给其传入数值直接计算归一化的结果
        norm_y = norm(bed['score'])
        colors = map_vir(norm_y)
    bed.loc[:,'length'] = abs(bed['end'] - bed['start'])
    bottom = list(range(intervals))  
    # broadcast
    if intervals < bed.shape[0]:
        repeat_times = (bed.shape[0] + intervals - 1) // intervals
        bottom = (bottom * repeat_times)[:bed.shape[0]]
    else:
        bottom = bottom[:bed.shape[0]]
    bed.loc[:, 'bottom'] = bottom

    ax.bar(x='start', width='length', height=1, bottom='bottom', align='edge', color= colors, edgecolor=None, data=bed, alpha=alpha)
    xlim_s = start
    xlim_e = end
    if needReverse == True:
         xlim_s = end
         xlim_e = start
    ax.set_xlim(xlim_s, xlim_e)
    if cname != None:
        sm = cm.ScalarMappable(norm=norm, cmap=map_vir)
        cax = mainAX.inset_axes([1.01, 0, 0.01, 1])
        cb = plt.colorbar(sm, ax=mainAX, cax=cax, label=score_label)
        cb.set_label(score_label, fontsize=score_label_size)


def _plot_bed_bar_l(ax, bed, start, end, needReverse, style='bar', color='tab:blue', alpha=1):
    if style == 'bar':
        ax.bar(x=bed['start'], width=bed['end']-bed['start'], height=bed['score'], color=color, alpha=alpha, align='edge')
    
    if style == 'line':
        ax.plot(bed['start'], bed['score'], color=color, alpha=alpha, solid_capstyle='butt')
    
    xlim_s = start
    xlim_e = end
    if needReverse == True:
         xlim_s = end
         xlim_e = start
    ax.set_xlim(xlim_s, xlim_e)
    
    for i in ['top','right', 'left']:
        ax.spines[i].set_color('none')
        ax.spines[i].set_linewidth(0)
    ax.spines["bottom"].set_color('black')
    ax.spines["bottom"].set_linewidth(1)
    #ax.tick_params(bottom =True,top=False,left=False,right=False)
    #ax.set_xticklabels("")
    #ax.set_yticklabels("")





import seaborn as sns
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.cm import get_cmap

def hex2rgb(value):
    # convert hex to rgb
    value = value.lstrip('#')
    lv = len(value)
    rgb = tuple(int(value[i:i + lv // 3], 16) for i in range(0, lv, lv // 3))
    return np.array([rgb[0]/255, rgb[1]/255, rgb[2]/255, 1])


def colorC(cname="RdBu_r", 
           bottom_color="#ffffff", 
           bad_color="white", 
           over_color="white", 
           under_color="white", 
           alpha=0):
    #cname = ["twilight_shifted", "jet", "RdBu_r", "RdGy_r", "BrBG_r", "hot_r", "Spectral_r"]
    #cname = ["terrain_r", "ocean", "gist_earth", "gist_stern_r", "tab20b", "twilight"]
    from matplotlib.colors import ListedColormap
    if isinstance(cname, str):
        #cmap=plt.get_cmap(cname)
        cmap = get_cmap(cname)
    else:
        cmap=cname

    cmap.set_bad(color=bad_color, alpha=alpha)
    cmap.set_over(color=over_color, alpha=alpha)
    cmap.set_under(color=under_color, alpha=alpha)

    if bottom_color==None:
        return cmap

    bottom_color = hex2rgb(bottom_color)
    newcolors = cmap(np.linspace(0, 1, 256))
    #white = np.array([1, 1, 1, 1])
    newcolors[0, :] = bottom_color
    newcmap = ListedColormap(newcolors)

    return newcmap

fruitpunch = sns.blend_palette(['white', 'red'], as_cmap=True)
#fruitpunch2 = sns.blend_palette(['white', 'blue'], as_cmap=True)
fruitpunch2 = sns.blend_palette(['white', 'purple'], as_cmap=True)





#-----------------------------------------------------------------




from trackc.pa.cmaps import (
    colorC,
    fruitpunch,
    fruitpunch2
)

from trackc.pa.palettes import (
    trackcl_9,
    trackcl_11,
    vega_10_trackc,
    default_20,
    zeileis_28,
    default_102
)



from trackc.tl._getRegionsCmat import extractContactRegions
from .mapc import mapC, getData2Map
from matplotlib.axes import Axes
from typing import Union, Sequence, Optional
from matplotlib.colors import Colormap
import pandas as pd
import numpy as np
import cooler
import seaborn as sns
fruitpunch = sns.blend_palette(['white', 'red'], as_cmap=True)

def _plot4C_line_bar(data,
                     ax,
                     color,
                     track_type,
                     logdata,
                     trim_range,
                     minrange,
                     maxrange,
                     ):
    data, maxrange, minrange = getData2Map(data, maxrange=maxrange, minrange=minrange, trim_range=trim_range, inplace=False)
    print("maxrange:", maxrange ,"minrange:",minrange)
    
    if logdata:
        data = np.log2(data)
    
    if track_type=='line':
        ax.plot(range(len(data[0,:])), data[0,:], color=color, solid_capstyle='butt')

    if track_type=='bar':
        ax.bar(x=range(len(data[0,:])), height=data[0,:], bottom=minrange, width=1, color=color, align='edge')
    return (data, maxrange, minrange)

def _get_pets(df, binsize=10000):
    pets = df.reset_index()
    pets_df = pd.DataFrame(index=range(pets['cbins'].sum()))
    pets_df['chrom'] = 'None'
    pets_df['bin_N'] = 0
    
    chroms = []
    bins = []
    
    for i, row in pets.iterrows():
        chroms.extend([row['chrom']] * row['cbins'])
         
        bin_s = int(row['fetch_start']/binsize)
        #if row['fetch_start']%binsize > 0:
        #    bin_s = bin_s

        bin_e = int(row['fetch_end']/binsize)
        if row['fetch_end']%binsize > 0:
            bin_e = bin_e + 1
        
        chrbins = list(range(bin_s, bin_e))
        
        if row['isReverse'] == True:
            chrbins = chrbins[::-1]
        bins.extend(chrbins)

    pets_df['chrom'] = chroms
    pets_df['chrbin'] = bins
    return pets_df


def virtual4C(ax: Optional[Axes] = None,
              clr: cooler.Cooler = None,
              balance: bool = False,
              #divisive_weights = None,
              target: Union[str, None] = None,
              contact_regions: Union[Sequence[str], str, None] = None, 
              track_type: Union[str, None] = 'line',
              color: Union[str, None] = 'tab:blue',
              cmap: Union[Sequence[Colormap], str, None] = fruitpunch,
              target_color: Union[str, None] = None,
              target_name: Union[str, None] = None,
              logdata: Union[Sequence[bool], bool] = False, 
              trim_range: float = 0.98,
              minrange: float = None,
              maxrange: float = None,
              label: Optional[str] = None,
              label_rotation: Union[int, None] = 0,
              label_fontsize: Optional[int] = 12,
              ):
    data = extractContactRegions(clr, balance=balance, row_regions=target, col_regions=contact_regions)
    
    if track_type in ['line', 'bar']:
        _, maxrange, minrange = _plot4C_line_bar(data=data.cmat,
                ax=ax,
                color=color,
                track_type = track_type,
                logdata=logdata, 
                trim_range=trim_range, 
                minrange=minrange, 
                maxrange=maxrange, 
                )
        
        cols = _get_pets(data.col_regions, clr.binsize)
        rows = _get_pets(data.row_regions, clr.binsize)
        cols['target'] = 0
        cols.loc[cols['chrbin'].isin(rows['chrbin']), 'target'] = 1
        targets_point = cols.query('target==1')
        target_bar = ax.bar(x=targets_point.index[0], height=maxrange, 
               bottom=minrange, width=1, color=target_color, align='edge', 
               label=target_name)
        #ax.bar_label(target_bar, label_type='edge')
        ax.text(targets_point.index[0], minrange+(maxrange-minrange)/2, target_name, va='center', ha='right', rotation=90)
        ax.set_ylim(minrange, maxrange)

    if track_type == 'heatmap':
        mapC(mat=data.cmat, cmap=cmap, logdata=logdata, 
             trim_range=trim_range, minrange=minrange, maxrange=maxrange, 
             ax=ax, map_type='square', ax_on=False)
        ax.set_aspect(aspect='auto')
    
    
    

    ax.tick_params(bottom =False,top=False,left=False,right=True)
    ax.yaxis.tick_right()
    #ax.set_xticklabels("")
    ax.set_xticklabels("")
    ax.set_xlim(0, data.cmat.shape[1]-1)
    ax.set_ylabel(label, fontsize=label_fontsize, rotation=label_rotation, ha='right', va='center')







from matplotlib.axes import Axes
#import pyBigWig
from typing import Union, Optional, Sequence, Any, Mapping, List, Tuple, Callable
import pandas as pd
from trackc.tl._getRegionsCmat import GenomeRegion

def _make_multi_region_ax(ax, lineGenomeRegions):
    lineGenomeRegions['len'] = lineGenomeRegions['fetch_end']-lineGenomeRegions['fetch_start']
    lineGenomeRegions['ax_ratio'] = lineGenomeRegions['len']/lineGenomeRegions['len'].sum()
    lineGenomeRegions['ax_x'] = lineGenomeRegions['ax_ratio'].cumsum(axis=0) - lineGenomeRegions['ax_ratio']
    axs = [ax.inset_axes([row['ax_x'], 0, row['ax_ratio'], 1]) for i, row in lineGenomeRegions.iterrows()]
    for axi in axs:
        axi.axis('off')
    
    return axs

def bw_track(bw,
             ax: Optional[Axes] = None,
             regions: Union[Sequence[str], str, None] = None, 
             binsize: Optional[int] = 50000,
             averagetype: Union[str, None] = 'mean',
             ymin: Optional[float] = None,
             ymax: Optional[float] = None,
             color: Union[Sequence[str], None] = '#827DBB',
             invert_y: Optional[bool] = False,
             label: Optional[str] = None,
             label_rotation: Union[int, None] = 0,
             label_fontsize: Optional[int] = 12,
             tick_fontsize: Optional[int] = 8,
             tick_fl: Optional[str] ='%0.2f', 
            ):
    """\
    Plot multi-regions bigwig signal tracks.
    
    Parameters
    ----------
    ax 
        ``cooler.Cooler``: cool format Hi-C matrix (https://github.com/open2c/cooler)
    ylabel
        ``str``: The ``'balance'`` parameters of ``coolMat.matrix(balance=False).fetch('chr6:119940450-123940450')``
    regions: bool, optional
        Force balancing weights to be interpreted as divisive (True) or
        multiplicativ

    binsize
        ``chrom region`` list: or ``chrom region`` or None. 
        The subset matrix row genome regions
        eg. ``"chr6:1000000-2000000"``, eg. ``["chr6:1000000-2000000", "chr3:5000000-4000000", "chr5"]``
        The start can be larger than the end (eg. ``"chr6:2000000-1000000"``), 
            which means you want to get the reverse region contact matrix

    averagetype
        ``chrom region`` list: or ``chrom region`` or None. 
        The subset matrix col genome regions, default is ``None``, which means the sample region as ``row_regions``
    """
    if isinstance(regions, list):
        line_GenomeRegions = pd.concat([GenomeRegion(i).GenomeRegion2df() for i in regions])
    else:
        line_GenomeRegions = GenomeRegion(regions).GenomeRegion2df()

    axs = _make_multi_region_ax(ax, line_GenomeRegions)
    line_GenomeRegions = line_GenomeRegions.reset_index()

    if isinstance(color, list)==False:
        color = [color]
    if len(color) < line_GenomeRegions.shape[0]:
        repeat_times = (line_GenomeRegions.shape[0] + len(color) - 1) // len(color)
        color = (color * repeat_times)[:line_GenomeRegions.shape[0]]
    
    min_y = 0
    max_y = 0
    
    for i, row in line_GenomeRegions.iterrows():    
        bins = int(row['len']/binsize)
        plot_list = bw.stats(row['chrom'], row['fetch_start'], row['fetch_end'], type=averagetype, nBins=bins)
        plot_list = [0 if v is None else v  for v in plot_list]

        axs[i].bar(x=range(0, bins), height=plot_list, width=1, bottom=[0]*(bins),color=color[i],align="edge",edgecolor=color[i])    
        
        right, left = bins, 0
        if row['isReverse'] == True:
            left, right = bins, 0
        axs[i].set_xlim(left, right)
        
        if min_y < min(plot_list):
            min_y = min(plot_list)
            
        if max_y < max(plot_list):
            max_y = max(plot_list)
        
    if ymin == None:
        ymin = min_y
    if ymax == None:
        ymax = max_y
        
    if invert_y == True:
        ymin = max_y
        ymax = min_y

    for axi in axs:
        axi.set_ylim(ymin, ymax)
        
    ax.set_ylim(ymin, ymax)
    
    va = 'top'
    if invert_y == True:
        va='bottom'
    ax.text(0, ymax, " [{0}, {1}]".format(tick_fl % min_y, tick_fl % ymax), verticalalignment=va, fontsize=tick_fontsize)
    
    ax.set_ylabel(label, fontsize=label_fontsize, rotation=label_rotation, 
                  horizontalalignment='right', verticalalignment='center')
     
    spines = ['top', 'bottom', 'left', 'right']
    if invert_y == True:
        del spines[0]
    else:
        del spines[1]
    for i in spines:
        ax.spines[i].set_visible(False)
    ax.set_xticks([])
    ax.set_xticklabels('')
    ax.set_yticks([])
    ax.set_yticklabels('')
    
def bw_compartment(compartment_bw, ax, chrom, start, end, ylabel, xticklabel=False, Acolor="#3271B2", Bcolor="#FBD23C", binsize=100000):
    chrsize = compartment_bw.chroms()[chrom]
    xbins = int(chrsize/binsize)
    if chrsize % binsize > 0:
        xbins = xbins + 1
        
    plot_list = compartment_bw.stats(chrom, 0, chrsize, nBins=xbins)
    mat=pd.DataFrame({"pc1":plot_list})
    mat["start"] = mat.index * binsize
    mat["width"] = binsize
    mat.loc[mat.index[-1],"length"] = chrsize % binsize
    mat.loc[mat[np.isnan(mat.pc1)].index, "pc1"] = 0
    
    plus = mat[mat['pc1']>0]
    minux = mat[mat['pc1']<=0]
 
    ax.bar(x=list(plus["start"]), height=plus['pc1'], width=plus["width"], bottom=[0]*(plus.shape[0]),color=Acolor,align="edge",edgecolor=Acolor,label="A")
    ax.bar(x=list(minux["start"]), height=minux['pc1'], width=minux["width"], bottom=[0]*(minux.shape[0]),color=Bcolor,align="edge",edgecolor=Bcolor,label="B")
    #ax.bar(0,height=0,color="#E27678",align="edge",edgecolor="#E27678",label="A2B")
    #ax.bar(0,height=0,color="#85AFBD",align="edge",edgecolor="#85AFBD",label="B2A")

    #ax.set_xlim(0, chrsize)
    ax.grid(False)
    ax.tick_params(bottom =False,top=False,left=True,right=False) #去掉tick线
    ax.spines['left'].set_color('k')
    ax.spines['left'].set_linewidth(1)
    ax.spines['right'].set_color('none')
    ax.spines['top'].set_color('none')
    ax.spines['bottom'].set_color('none')
    ax.plot([0, chrsize], [0,0], '-', label='', linewidth=1, color='black', solid_capstyle='butt')
    #ax.set_yticklabels('')
    ax.set_ylabel(ylabel, fontsize=10, rotation='horizontal', horizontalalignment='right',verticalalignment='center')
    #ax.set_ylim([-1,1])
    #ax.set_yticks([-1, 1])
    ax.set_yticklabels([-1, 1], fontsize=8)
    if xticklabel==False:
        ax.set_xticklabels('')
    
    ax.set_xlim(start, end)

    



import numpy as np
import pandas as pd
from matplotlib.colors import Colormap
from mpl_toolkits.axes_grid1.inset_locator import inset_axes
from matplotlib.axes import Axes
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm #,CenteredNorm, SymLogNorm,PowerNorm, Normalize  
from typing import Union, Optional, Sequence, Any, Mapping, List, Tuple, Callable
from trackc.pa import fruitpunch, fruitpunch2
import os
basedir = os.path.abspath(os.path.dirname(__file__))

ColorLike = Union[str, Tuple[float, ...]]

def _check_na_color(
    na_color: Optional[ColorLike], *, img: Optional[np.ndarray] = None
) -> ColorLike:
    if na_color is None:
        if img is not None:
            na_color = (0.0, 0.0, 0.0, 0.0)
        else:
            na_color = "lightgray"
    return na_color

def getData2Map(mat, maxrange=None, minrange=None, trim_range=0.99, inplace=False):
    mat = mat.astype(float)
    #if logdata: mat = np.log2(mat)   
    mat[mat == np.inf] = 0.0
    mat[mat == -np.inf] = 0.0
    mat[mat == 0.0] = np.nan    
    
    df = pd.DataFrame(np.ravel(mat))
    df = df[df != np.nan]
    
    if np.nanmax(mat)>0:
        if trim_range <1 and maxrange==None and minrange==None:
            xmaxrange = np.nanpercentile(abs(df), trim_range*100)
            xminrange = np.nanpercentile(abs(df), 100 - trim_range*100)
            print("no max min range")
        else:
            if maxrange==None:
                xmaxrange = abs(df).max()[0]
            else:
                xmaxrange = maxrange
            if minrange==None:
                xminrange = abs(df).min()[0]
            else:
                xminrange = minrange

        if inplace==True:
            mat[(mat<=xminrange) & (mat>0)] = xminrange    
            mat[(mat>=xmaxrange) & (mat>0)] = xmaxrange
            mat[(mat>= -xminrange) & (mat<0)] = -xminrange
            mat[(mat<= -xmaxrange) & (mat<0)] = -xmaxrange

        return mat, xmaxrange, xminrange
    else:
        print ("Warning: max data <0, no data to plot")


def _plot_pcolormesh( 
    mat: np.ndarray,
    ax: Axes,
    norm=None,
    clim=None,
    cmap: Union[Colormap, str, None] = None,
    trans_ax: bool=False
    ):

    N = mat.shape[1]
    # Transformation matrix for rotating the heatmap.
    A = np.array([(y, x) for x in range(N, -1, -1) for y in range(N + 1)])
    t = np.array([[1,0.5], [-1,0.5]])
    A = np.dot(A, t)
  
    C = mat
    X = A[:, 1].reshape(N + 1, N + 1)
    Y = A[:, 0].reshape(N + 1, N + 1)

    if trans_ax==False:
        caxes = ax.pcolormesh(X, Y, np.flipud(C), cmap=cmap, edgecolor='none', norm=norm, clim=clim, snap=True, linewidth=.001)
    else:
        caxes = ax.pcolormesh(Y, X, np.flipud(C), cmap=cmap, edgecolor='none', norm=norm, clim=clim, snap=True, linewidth=.001)
    
    return caxes

def _mapC_triview(
    mat: np.ndarray,
    ax: Optional[Axes] = None,
    cmap: Union[Colormap, str, None] = fruitpunch,
    label: Union[str, None] = None,
    label_fontsize: int = 10,
    label_color: Union[str, None] = 'k',
    logdata: bool = False, 
    trim_range: float = 0.98,
    maxrange: float = None,
    minrange: float = None,
    height: int = 0,
    trans_ax: bool = False,
    map_type: Union[str, None] = 'triangle',
    map_order: int = 0,
    symmetric: bool = False,
    k = 1
    ): 
    """\
    Draw triangle view of the C data

    Parameters
    mat:

    cmap
    map_type
        contact heatmap type, can be one of ``['square', 'triangle', 'rectangle']``,
        if select ``rectangle`` the ``mapHeight`` parameter should be set.
    mapHeight: int or None
        Parameter for ``map_type: rectangle'``, ``map_type: triangle'`` also can be set.
        it means the heapmap hight bin number

    map_order: int
        0: np.triu()
        1: np.tril()
    k: int
        ``np.tril`` or ``np.triu`` ``k`` parameter

    """
    mat, maxrange, minrange = getData2Map(mat, maxrange=maxrange, minrange=minrange, trim_range=trim_range, inplace=False)
    print("maxrange:", maxrange ,"minrange:",minrange)

    if symmetric==False:
        if map_order==0:
            mat = np.triu(mat, k = k)
            mat[mat==0.0] = np.nan
        elif map_order==1:
            mat = np.tril(mat, k = k)
            mat[mat==0.0] = np.nan
        else:
            pass
    
    norm = None
    if logdata:
        norm=LogNorm()
    
    clim = (minrange, maxrange)
    if map_type=='square':
        im = ax.matshow(mat, cmap=cmap, norm=norm, clim=clim)
    else:
        if map_type=='triangle':
            im = _plot_pcolormesh(mat=mat, ax=ax, norm=norm, clim=clim, cmap=cmap, trans_ax=trans_ax)
 
        if map_type=='rectangle':
            im = _plot_pcolormesh(mat=mat, ax=ax, norm=norm, clim=clim, cmap=cmap, trans_ax=trans_ax)

    _mapC_colorbar(ax, im, trans_ax, map_order, height, map_type)

    return im

def _mapC_label(ax, label, trans_ax, map_order, map_type, height, fontsize= 10, color='k', Pair_mat=False):
    xmin, xmax = ax.get_xlim() 
    ymin, ymax = ax.get_ylim() 

    x, y = xmax, ymax
    verticalalignment = 'top'
    horizontalalignment = 'right'

    if map_type == "square":
        if map_order==0:
            x, y = xmax, ymax
        else:
            x, y = xmin, ymin
            verticalalignment = 'bottom'
            horizontalalignment ='left'
    else: # triangle, rectangle
        if Pair_mat==False:
            if trans_ax == True:
                ax.set_xlabel(label, fontsize=fontsize, color=color)
            else:
                ax.set_ylabel(label, fontsize=fontsize, color=color)
            
            return
        
        if trans_ax == True:
            y = ymax
            if map_order==0:
                x = xmax
            else:
                x = xmin
                horizontalalignment ='left'
        else:
            x = xmin
            horizontalalignment ='left'
            if map_order==1:
                y = ymin
                verticalalignment = 'bottom'

    bbox_props = dict(boxstyle="round", fc="w", alpha=0.9, pad=0, ec='none')

    ax.text(x, y, 
            label, 
            verticalalignment = verticalalignment,
            horizontalalignment = horizontalalignment,
            #transform = ax.transAxes,
            color =color, 
            fontsize = fontsize,
            bbox = bbox_props
            )

def _mapC_colorbar(ax, im, trans_ax, map_order, height, map_type):
    x0, y0, x0_width, y0_height = 0, 1.015, 1, 0.015
    if trans_ax == False:
        x0, y0, x0_width, y0_height = 1.015, 0, 0.015, 1
        if map_order in [0, 1]:
            y0_height = 0.45
            if height > 0 :
                x0_width = 0.012
            else:
                x0_width = 0.04
            if map_order==0:
                y0 = 0.55
            else:
                y0 = 0
        else:
            x0_width = 0.04
            if height > 0 :
                x0_width = 0.012
    else:
        if map_order in [0, 1]:
            x0_width = 0.45
            if height > 0 :
                y0_height = 0.012
            else:
                y0_height = 0.04
            if map_order==0:
                x0 = 0.55
            else:
                x0 = 0
        else:
            y0_height = 0.04
            if height > 0 :
                y0_height = 0.012

    orientation = 'vertical'
    if trans_ax:
        orientation = 'horizontal'
    
    cax = ax.inset_axes([x0, y0, x0_width, y0_height])
    cbar = plt.colorbar(im, ax=ax, cax=cax, orientation=orientation)
    if trans_ax:
        cbar.ax.xaxis.tick_top()
        cbar.ax.tick_params(axis='x', labelrotation=90)
    else:
        cbar.ax.yaxis.tick_right()
    #cbar.set_label(label, fontsize=8)

def _paraPair(para):
    if isinstance(para, list)==False:
        return([para, para])
    else:
        if len(para)==1:
            para[1] = para[0]
        else:
            return para
    return para

def mapC(
        ax: Optional[Axes] = None,
        mat: Union[np.ndarray, None]  = None,
        mat2: Union[np.ndarray, None] = None,
        
        cmap: Union[Sequence[Colormap], Sequence[str], Colormap, str, None] = [fruitpunch, 'YlOrRd'],
        label: Union[Sequence[str], str, None] = None,
        label_fontsize: Union[Sequence[int], int] = 10,
        label_color: Union[Sequence[str], str, None] = 'k',
        
        logdata: Union[Sequence[bool], bool] = False, 
        maxrange: Union[Sequence[float], float]=None,
        minrange: Union[Sequence[float], float]=None,
        trim_range: Union[Sequence[float], float]=0.99,
        
        map_type: Union[str, None] = 'triangle',
        height: int = 0,
        trans_ax: bool = False,
        symmetric: bool = False,
        ax_on: bool =True,
        aspect: Union[str, float]='auto'
        ):
    """\
    Draw triangle view of the C data

    Parameters
    mat:

    cmap
    map_type
        contact heatmap type, can be one of ``['square', 'triangle', 'rectangle']``,
        if select ``rectangle`` the ``mapHeight`` parameter should be set.
    mapHeight: int or None
        Parameter for ``map_type: rectangle'``, ``map_type: triangle'`` also can be set.
        it means the heapmap hight bin number
    symmetric: bool
        If there is one of ``mat`` and ``mat2`` para is None, 
        value ``True`` means the  symmetric heatmap
    aspect: 'auto' or 1
    """

    #cmap = copy(get_cmap(cmap))
    #cmap.set_bad(na_color)
    #cmap2 = copy(get_cmap(cmap2))
    #cmap2.set_bad(na_color)
    cmap = _paraPair(cmap)
    label = _paraPair(label)
    label_fontsize = _paraPair(label_fontsize)
    label_color = _paraPair(label_color)
    logdata = _paraPair(logdata)
    maxrange = _paraPair(maxrange)
    minrange = _paraPair(minrange)
    trim_range = _paraPair(trim_range)

    k = 1
    k2 = -1

    if isinstance(mat, np.ndarray)==False:
        k2 = 0
    if isinstance(mat2, np.ndarray)==False:
        k = 0

    #ax2 = ax
    Pair_mat = False
    ax2 = ax.inset_axes([0, 0, 1, 1], facecolor='none')
    ax2.set_zorder(1)
    ax2.set_xticklabels([])
    ax2.set_xticks([])
    ax2.set_yticklabels([])
    ax2.set_yticks([]) 


    if isinstance(mat, np.ndarray) and isinstance(mat2, np.ndarray):
        #ax2 = ax.inset_axes([0, 0, 1, 1], facecolor='none')
        #ax2.set_zorder(-1)
        symmetric = False
        Pair_mat = True

    if isinstance(mat, np.ndarray):
        im = _mapC_triview(
            mat=mat,
            ax=ax,
            cmap=cmap[0],
            label=label[0],
            logdata=logdata[0], 
            maxrange=maxrange[0],
            minrange = minrange[0],
            trim_range=trim_range[0],
            height=height,
            trans_ax=trans_ax,
            map_type=map_type,
            map_order=0,
            symmetric=symmetric,
            k = k
        )

    if isinstance(mat2, np.ndarray):
        #ax2 = ax.inset_axes([0, 0, 1, 1], facecolor='none')
        #ax2.set_zorder(-1)
        
        im2 = _mapC_triview(
            mat=mat2,
            ax=ax2,
            cmap=cmap[1],
            label=label[1],
            logdata=logdata[1], 
            maxrange=maxrange[1],
            minrange = minrange[1],
            trim_range=trim_range[1],
            height=height,
            trans_ax=trans_ax,
            map_type=map_type,
            map_order=1,
            symmetric=symmetric,
            k = k2
        )
        
    xylim_conf = pd.read_table(os.path.join(basedir,'mapc_xylim.txt'), header=0)

    if isinstance(mat, np.ndarray) and isinstance(mat2, np.ndarray):
        set_xylim(ax, xylim_conf, map_type, 0, symmetric, trans_ax, Pair_mat, mat, height)
        set_xylim(ax2, xylim_conf, map_type, 1, symmetric, trans_ax, Pair_mat, mat2, height)

        ax.set_aspect(aspect)
        ax2.set_aspect(aspect)
        
    elif isinstance(mat, np.ndarray):
        set_xylim(ax, xylim_conf, map_type, 0, symmetric, trans_ax, Pair_mat, mat, height)
        ax.set_aspect(aspect)
    elif isinstance(mat2, np.ndarray):
        set_xylim(ax, xylim_conf, map_type, 1, symmetric, trans_ax, Pair_mat, mat2, height)
        set_xylim(ax2, xylim_conf, map_type, 1, symmetric, trans_ax, Pair_mat, mat2, height)
        ax.set_aspect(aspect)
        ax2.set_aspect(aspect)
    else:
        pass

    if ax_on==False:
        ax.axis('off')
        ax2.axis('off')

    if isinstance(mat, np.ndarray):
        _mapC_label(ax, label[0], trans_ax, 0, map_type, height, fontsize=label_fontsize[0], color=label_color[0], Pair_mat=Pair_mat)
    if isinstance(mat2, np.ndarray):
        _mapC_label(ax2, label[1], trans_ax, 1, map_type, height, fontsize=label_fontsize[1], color=label_color[1], Pair_mat=Pair_mat)

def set_xylim(axs, conf, map_type, map_order, symmetric, trans_ax, pair_mat, matx, height):
    conf = conf[(conf['map_type']==map_type) & (conf['map_order']==map_order) & (conf['symmetric']==symmetric) & (conf['trans_ax']==trans_ax) & (conf['pair_mat']==pair_mat)]
    if height == 0:
        height = matx.shape[0]
    
    conf_values = {'mat_w': matx.shape[0], 'mat_h': matx.shape[1], 'height': height, 'height_m': -height, 
                   'mat_w_height': matx.shape[0]-height, '0':0}
    
    if conf.shape[0] == 0:
        print('type not set xylim')
    else:
        ix = conf.index[0]
        xmin = conf_values[conf.loc[ix, 'xlim_left']]
        xmax = conf_values[conf.loc[ix, 'xlim_right']]
        ymin = conf_values[conf.loc[ix, 'ylim_bottom']]
        ymax = conf_values[conf.loc[ix, 'ylim_top']]
        axs.set_xlim([xmin, xmax])
        axs.set_ylim([ymin, ymax])

    
    axs.set_xticklabels([])
    axs.set_xticks([])
    axs.set_yticklabels([])
    axs.set_yticks([]) 
    
    

def _plot_heatmap_triangle_xticks(ax, regin1_binN, regin2_binN, chrom1, start1, end1, chrom2, start2, end2, showXticks):
    #ax.set_xticks([])
    ax.set_xticks([0, regin1_binN, regin1_binN + regin2_binN])
    ax.set_xticklabels([start1, str(end1) + " " + str(start2), end2])
    
    ax.set_yticklabels([])
    ax.set_yticks([])
    
    ax.xaxis.tick_top()
    ax.spines['top'].set_color('k')
    ax.spines['right'].set_color('none')
    ax.spines['bottom'].set_color('none')
    ax.spines['left'].set_color('none')

   
def _colorbar_triangle(axm,im,ymax):
    height="2%"
    width="10%"
    if ymax!=None:
        height="6%"
    axins1 = inset_axes(axm, width=width, height=height, loc=4, bbox_to_anchor=(-0.1, 0.7, 1.2, 2.9), bbox_transform=axm.transAxes,)
    cbar=plt.colorbar(im, cax=axins1, orientation='horizontal')
    cbar.ax.xaxis.tick_top()
    cbar.ax.spines['top'].set_color('none')
    cbar.ax.spines['right'].set_color('none')
    cbar.ax.spines['bottom'].set_color('none')
    cbar.ax.spines['left'].set_color('none')



        




# %% [markdown]
# # Custom Notebook Formats
#
# By default, Jupyter notebooks are stored in files with the suffix `.ipynb`,
# which use the JSON format for storage.
#
# However, there are libraries available which allow storing notebooks
# in different formats, using different file suffixes.
#
# To use a custom notebook format in `nbsphinx`, you can specify the
# `nbsphinx_custom_formats` option in your `conf.py` file.
# You have to provide the file extension
# and a conversion function that takes the contents of a file (as a string)
# and returns a Jupyter notebook object.
#
# ```python
# nbsphinx_custom_formats = {
#     '.mysuffix': 'mylibrary.converter_function',
# }
# ```
#
# The converter function can be given as a string (recommended)
# or as a function object.
#
# If a conversion function takes more than a single string argument,
# you can specify the function name plus a dictionary with keyword arguments
# which will be passed to the conversion function in addition to the
# file contents.
#
# ```python
# nbsphinx_custom_formats = {
#     '.mysuffix': ['mylibrary.converter_function', {'some_arg': 42}],
# }
# ```
#
# You can of course use multiple formats
# by specifying multiple conversion functions.

# %% [markdown]
# ## Example: Jupytext
#
# One example for a library which provides a custom conversion function is
# [jupytext](https://github.com/mwouts/jupytext),
# which allows storing the contents of Jupyter notebooks in
# Markdown and R-Markdown, as well as plain Julia, Python and R files.
#
# Since its conversion function takes more than a single string argument,
# we have to pass a keyword argument, e.g.:
#
# ```python
# nbsphinx_custom_formats = {
#     '.Rmd': ['jupytext.reads', {'fmt': 'Rmd'}],
# }
# ```

# %% [markdown]
# This very page is an example of a notebook stored in the
# `py:percent` format
# (see [docs](https://jupytext.readthedocs.io/en/latest/formats.html#the-percent-format)):

# %%
# !head -20 custom-formats.pct.py

# %% [raw] raw_mimetype="text/restructuredtext"
# To select a suitable conversion function,
# we use the following setting in :download:`conf.py`:
#
# .. literalinclude:: conf.py
#     :language: python
#     :start-at: nbsphinx_custom_formats
#     :lines: -4
#     :emphasize-lines: 2

# %% [markdown]
# Another example is [this gallery example page](gallery/due-rst.pct.py).



import matplotlib.pyplot as plt
from typing import List, Union
from matplotlib.axes import Axes

def make_spec(width: float = 10,
              height: float = 3,
              height_ratios: Union[List[float], None] = None,
              width_ratios: Union[List[float], None] = None,
              hspace: float = 0,
              wspace: float = 0,

              ):
    """\
    """
    
    nfig = 0
    nrows = 1
    ncols = 1
    ratios = height_ratios
    if height_ratios != None and isinstance(height_ratios, list):
        nfig = len(height_ratios)
        width_ratios = None
        nrows = nfig

    if width_ratios != None and isinstance(width_ratios, list):
        nfig = len(width_ratios)
        height_ratios = None
        ncols = nfig
        ratios = width_ratios

    if nfig <= 0:
        fig, ax = plt.subplots(1, 1, figsize=(width, height) )
        return fig, ax
    
    fig = plt.figure(figsize=(width, height))
    spec = fig.add_gridspec(nrows=nrows, ncols=ncols,
                           left=None, bottom=None, right=None, top=None,
                           wspace=wspace, hspace=hspace, 
                           width_ratios=width_ratios, 
                           height_ratios=height_ratios)
    
    axs = [None] * nfig
    for i, v in enumerate(ratios):
        axs[i] = fig.add_subplot(spec[i])

    return fig, axs

# fig, axs = make_spec(height_ratios=[1,1.5,2.4], hspace=0.1)

class mortise:
    bottom = 0
    top = 0
    track_height = 0
    hspace = 0
    ax = None
    def __init__(self, ax_ref: Axes, bottom: float, track_height: float, hspace: float):
        self.ax = ax_ref.inset_axes([0, bottom, 1, track_height])
        self.bottom = bottom
        self.top = bottom + track_height
        self.track_height = track_height
        self.hspace = hspace
    
class tenon:
    mortises = []
    fig = None
    ax = None
    def __init__(self, width=7, height=1 ):
        fig, ax = plt.subplots(1, 1, figsize=(width, height) )
        ax.set_axis_off()
        self.fig = fig
        self.ax = ax
        
    def show(self):
       for i, v in enumerate(self.mortises):
           #v.ax.set_ylabel('mortises[{0}]'.format(i), rotation=0)
           v.ax.text(0.5, 0.5, '.axs({0})'.format(i), ha='center', va='center')
           v.ax.set_xticklabels([])
           v.ax.set_xticks([])
       self.fig.show()

    def add(self, pos='top', height=1, hspace=0):
        if len(self.mortises) != 0:
            ix = -1
            if pos=="top":
                ix = 0 
            ref_lego_ax = self.mortises[ix]
        else:
            ref_lego_ax = mortise(self.ax, 0, height, 0)
            self.mortises = [ref_lego_ax]
            return
        
        bottom = ref_lego_ax.bottom - hspace - height
        if pos == 'top':
            bottom = ref_lego_ax.top + hspace

        lego_ax_demo = mortise(self.ax, bottom, height, hspace)
        if pos == 'top':
            self.mortises.insert(0, lego_ax_demo)
        else:
            self.mortises.append(lego_ax_demo)

    def remove(self, pos='top'):
        if len(self.mortises) >=1:
            if pos == 'top':
                self.mortises.remove(self.mortises[0])
            elif pos == 'bottom':
                self.mortises.remove(self.mortises[-1])
            else:
                pass
    def axs(self, index):
        return self.mortises[index].ax

def savefig(outfile):
    plt.tight_layout()
    plt.savefig(outfile, bbox_inches='tight')




from trackc.pl.scale import multi_scale_track, scale_track
from trackc.pl.bed import bed_track
from trackc.pl.bigwig import bw_track, bw_compartment
from trackc.pl.links import links_track
from trackc.pl.Virtual4C import virtual4C
from trackc.pl.mapc_markline import mapc_markline
from trackc.pl.gene import gene_track
from trackc.pl.mapc import (
    mapC,
    getData2Map
)
from trackc.pl.zoomin import zoomin

__doc__ = """\
Plotting API
============

.. currentmodule:: trackc

.. note::
   See the :ref:`API` section for all important plotting configurations.

.. _pl-generic: 

Generic
-------

.. autosummary::
   :toctree: .
   pl.getData2Map
   pl.mapC
   pl.gene_track
   pl.mapc_markline
   pl.virtual4C
   pl.links_track
   pl.bw_track
   pl.bw_compartment
   pl.bed_track
   pl.multi_scale_track
"""



from matplotlib.axes import Axes
from typing import Union, Optional, Sequence
import pandas as pd
import numpy as np
from trackc.tl._getRegionsCmat import GenomeRegion
from .bigwig import _make_multi_region_ax


def _two_degree_bc(x_l=10, x_r=90, y_lr=0, y2=10, dots_num=100): 
    """
        bezier curve for loop links
    """

    xt = []
    yt = []
    x_mid = (x_l + x_r)/2
    x_dots12 = np.linspace(x_l, x_mid, dots_num)
    y_dots12 = np.linspace(y_lr, y2, dots_num)
    x_dots23 = np.linspace(x_mid, x_r, dots_num)
    y_dots23 = np.linspace(y2, y_lr, dots_num)
    for i in range(dots_num):
        x = x_dots12[i] + (x_dots23[i]-x_dots12[i])*i / (dots_num-1)
        y = y_dots12[i] + (y_dots23[i]-y_dots12[i])*i / (dots_num-1)
        xt.append(x)
        yt.append(y)
    return (xt, yt)

def _plot_loop(ax, loop_df, color, max_extend, invert_y, start, end, left_anchor, right_anchor):
    if loop_df.shape[0] == 0:
        return
    
    top_y = 0

    for i, row in loop_df.iterrows():
        top = row['length']/max_extend
        if top < 0.5:
            top = 0.5
        elif top >0.8:
            top = 0.8
        else:
            pass
        
        xt, yt = _two_degree_bc(x_l=row[left_anchor], x_r=row[right_anchor], y_lr=0, y2=top, dots_num=100)
    
        ax.plot(xt, yt, color=color, linewidth=0.5, solid_capstyle='butt')
        if max(yt) > top_y:
            top_y = max(yt)
            
    ax.set_xlim(start, end)
    if invert_y==True:
        ax.set_ylim(0.5, 0)
    else:
        ax.set_ylim(0, 0.5)

def links_track(
        ax: Optional[Axes] = None,
        links_df: pd.DataFrame = None, 
        regions: Union[Sequence[str], str, None] = None,
        links_type: Union[str, None] = 'loop',
        color: Union[Sequence[str], None] = '#66AC84',
        maxrange: Union[int, None] = 3000000,
        invert_y: Optional[bool] = False,
        anchor: Union[str, None] = 'inside',

        label: Optional[str] = None,
        label_rotation: Union[int, None] = 0,
        label_fontsize: Optional[int] = 12,
        ):
    """\
    Plot multi-regions loop or TAD links.
    
    Parameters
    ----------
    loop_bed
        ``pd.DataFrame``: the file format for links is (tab separated), column names:
            chr1 x1 x2 chr2 y1 y2 (score ...)
        The score field is optional
        The fields after the y2 score will be ignored, score column is useless right now
        for example:
            chr1 100 200 chr1 250 300 1

    ax 
        ``cooler.Cooler``: cool format Hi-C matrix (https://github.com/open2c/cooler)
    ylabel
        ``str``: The ``'balance'`` parameters of ``coolMat.matrix(balance=False).fetch('chr6:119940450-123940450')``
    regions:
        bool, optional
        Force balancing weights to be interpreted as divisive (True) or
        multiplicativ
    links_type: 
        ``str``: links type, either 'loop' or 'triangle'
    
    
    anchor:
        ``str``: links type, either 'loop' or 'triangle'
            the link coordinates type for loop and triangle, by default: inside
                inside: link x2 and y1
                mid: link the middle of x1 and x2 and the middle of y1 and y2, (x1+x2)/2 and (y1+y2)/2
                outside: link x1 and y2
    """

    if isinstance(regions, list):
        line_GenomeRegions = pd.concat([GenomeRegion(i).GenomeRegion2df() for i in regions])
    else:
        line_GenomeRegions = GenomeRegion(regions).GenomeRegion2df()

    axs = _make_multi_region_ax(ax, line_GenomeRegions)
    line_GenomeRegions = line_GenomeRegions.reset_index()

    if isinstance(color, list)==False:
        color = [color]
    if len(color) < line_GenomeRegions.shape[0]:
        repeat_times = (line_GenomeRegions.shape[0] + len(color) - 1) // len(color)
        color = (color * repeat_times)[:line_GenomeRegions.shape[0]]
    
    if anchor == "mid":
        left_anchor = 'left_mid'
        right_anchor = 'right_mid'

    if anchor == "inside":
        left_anchor = 'x2'
        right_anchor = 'y1'
    if anchor == "outside":
        left_anchor = 'x1'
        right_anchor = 'y2'

    max_extend = 0
    links_df_list = []

    links_df = links_df[['chr1', 'x1', 'x2', 'chr2', 'y1', 'y2']]
    links_df['chr1'] = links_df['chr1'].astype(str)
    links_df['chr2'] = links_df['chr2'].astype(str)

    for ix, row in line_GenomeRegions.iterrows(): 
        loop_bed_plot = links_df[links_df['chr1']==row['chrom']]
        

        loop_bed_plot['length'] = loop_bed_plot["y2"] - loop_bed_plot["x1"]
        
        loop_bed_plot = loop_bed_plot[((loop_bed_plot['x1'] >= row['fetch_start']) & (loop_bed_plot['x1'] <= row['fetch_end'])) | \
                                ((loop_bed_plot['x2'] >= row['fetch_start']) & (loop_bed_plot['x2'] <= row['fetch_end']) |\
                                (loop_bed_plot['y1'] >= row['fetch_start']) & (loop_bed_plot['y1'] <= row['fetch_end'])) | \
                                    ((loop_bed_plot['y2'] >= row['fetch_start']) & (loop_bed_plot['y2'] <= row['fetch_end']))]

        loop_bed_plot = loop_bed_plot[(loop_bed_plot["y2"] - loop_bed_plot["x1"])<maxrange] 
        #print('looop num:', loop_bed_plot.shape)
        if loop_bed_plot.shape[0] == 0:
            #for i in ['top', 'right', "left", "bottom"]:
            #    ax.spines[i].set_color('none')
            #    ax.spines[i].set_linewidth(0)
            next

        loop_bed_plot['left_mid'] = (loop_bed_plot['x1']+loop_bed_plot['x2'])/2
        loop_bed_plot['right_mid'] = (loop_bed_plot['y1']+loop_bed_plot['y2'])/2

        links_df_list.append(loop_bed_plot)
        if loop_bed_plot.shape[0] >0:
            if max_extend < max(loop_bed_plot['length']):
                max_extend = max(loop_bed_plot['length'])
   
    for ix, row in line_GenomeRegions.iterrows(): 
        if links_type == 'loop':
            _plot_loop(axs[ix], links_df_list[ix], color[ix], max_extend, invert_y, row['start'], row['end'], left_anchor, right_anchor)
        #if links_type == 'loop':
            #plot_triangle()

    ax.set_ylabel(label, fontsize=label_fontsize, rotation=label_rotation, 
                  horizontalalignment='right',verticalalignment='center')
    
    spines = ['top', 'bottom', 'left', 'right']
    if invert_y == True:
        del spines[0]
    else:
        del spines[1]
    for i in spines:
        ax.spines[i].set_visible(False)
    ax.set_xticks([])
    ax.set_xticklabels('')
    ax.set_yticks([])
    ax.set_yticklabels('')




import logging
import cooler
import pandas as pd
import numpy as np
from typing import Union, List, Sequence
import cooler

class GenomeRegion:
    region = None
    chrom = None
    start = None
    end = None
    fetch_start = None
    fetch_end = None
    length = None
    isReverse = False
    
    def __init__(self, region: str):
        self.region = region
        tmp = region.split(":")
        self.chrom = tmp[0]
        if len(tmp)==2:
            self.start = int(tmp[1].split("-")[0])
            self.end = int(tmp[1].split("-")[1])
            if self.start > self.end:
                self.isReverse = True
                self.fetch_start = self.end
                self.fetch_end = self.start
            else:
                self.fetch_start = self.start
                self.fetch_end = self.end
            self.length = abs(self.start - self.end)

    def fetchRegion(self):
        fetch_r = self.chrom
        if self.fetch_start != None:
            fetch_r = fetch_r + ":" + str(self.fetch_start) + "-" +  str(self.fetch_end)
        return fetch_r

    def GenomeRegion2df(self):
        """
        region4coolFetch = self.chrom + ":" + str(self.start) + '-' + str(self.end)
        if self.start == None:
            region4coolFetch = self.chrom
        else:
            if self.start > self.end:
                region4coolFetch = self.chrom + ":" + str(self.end) + '-' + str(self.start)
        """
        region4coolFetch = self.fetchRegion()

        df = pd.DataFrame(
            {'chrom':[self.chrom], 
            'start':[self.start],
            'end':[self.end],
            'isReverse':[self.isReverse],
            'fetch_start':[self.fetch_start],
            'fetch_end':[self.fetch_end],
            'region4coolFetch': [region4coolFetch]
            }, 
            index=[self.region])

        return df
        
class RegionsCmat:
    def __init__(self, cmat: np.array, row_regions: pd.DataFrame, col_regions: pd.DataFrame):
        self.cmat = cmat
        self.row_regions = row_regions
        self.col_regions = col_regions

# import extractContactRegions as subsetContactRegions

def extractContactRegions(
        clr: cooler.Cooler, 
        balance: bool = False,
        #divisive_weights = None,
        row_regions: Union[Sequence[str], str, None] = None,
        col_regions: Union[Sequence[str], str, None] = None, 
        ) -> RegionsCmat:
    """\
    Extract a set of regions matrix from the cool format Hi-C matrix.

    The extracted matrix will splice intra and Inter region interaction according to 
        the given order and direction of the regions.
    Parameters
    ----------
    coolMat 
        ``cooler.Cooler``: cool format Hi-C matrix (https://github.com/open2c/cooler)
    balance
        ``bool``: The ``'balance'`` parameters of ``coolMat.matrix(balance=False).fetch('chr6:119940450-123940450')``
    divisive_weights: bool, optional
        Force balancing weights to be interpreted as divisive (True) or
        multiplicative (False). Weights are always assumed to be
        multiplicative by default unless named KR, VC or SQRT_VC, in which
        case they are assumed to be divisive by default.

    row_regions
        ``chrom region`` list: or ``chrom region`` or None. 
        The subset matrix row genome regions
        eg. ``"chr6:1000000-2000000"``, eg. ``["chr6:1000000-2000000", "chr3:5000000-4000000", "chr5"]``
        The start can be larger than the end (eg. ``"chr6:2000000-1000000"``), 
            which means you want to get the reverse region contact matrix

    col_regions
        ``chrom region`` list: or ``chrom region`` or None. 
        The subset matrix col genome regions, default is ``None``, which means the sample region as ``row_regions``
        
    Returns:
        :class:`~trackc.RegionsCmat`
            row_regions and col_regions contact matrix object
    """
    # -------
    if isinstance(row_regions, list):
        row_GenomeRegions = pd.concat([GenomeRegion(i).GenomeRegion2df() for i in row_regions])
    else:
        row_GenomeRegions = GenomeRegion(row_regions).GenomeRegion2df()

    if col_regions == None:
        col_GenomeRegions = row_GenomeRegions.copy()
    else:
        if isinstance(col_regions, list):
            col_GenomeRegions = pd.concat([GenomeRegion(i).GenomeRegion2df() for i in col_regions])
        else:
            col_GenomeRegions = GenomeRegion(col_regions).GenomeRegion2df()
    
    # ------
    region_mat_dic = {}
    for _, row_row in row_GenomeRegions.iterrows():
        for _, col_row in col_GenomeRegions.iterrows():
            row_col_region_cmat = clr.matrix(balance=balance).fetch(row_row['region4coolFetch'], col_row['region4coolFetch'])
            if row_row['isReverse'] == True:
                row_col_region_cmat = np.flip(row_col_region_cmat, 0)
            if col_row['isReverse'] == True:
                row_col_region_cmat = np.flip(row_col_region_cmat, 1)

            region_mat_dic["{0}__{1}".format(row_row["region4coolFetch"], col_row["region4coolFetch"])] = row_col_region_cmat

    vstack_list = [None]*row_GenomeRegions.shape[0]
    hstack_list = [None]*col_GenomeRegions.shape[0]
    
    for i, row_region in enumerate(row_GenomeRegions["region4coolFetch"].to_list()):
        for ii, col_region in enumerate(col_GenomeRegions["region4coolFetch"].to_list()):
            hstack_list[ii] = region_mat_dic["{0}__{1}".format(row_region, col_region)]
        
        vstack_list[i] = np.hstack(tuple(hstack_list))
                          
    cMat = np.vstack(tuple(vstack_list))

    #### 
    row_GenomeRegions['cbins'] = 0
    col_GenomeRegions['cbins'] = 0

    row_index = row_GenomeRegions.index.to_list()
    col_index = col_GenomeRegions.index.to_list()

    #print(row_index)
    #print(col_index)

    for i in row_index:
        for ii in col_index:
            cmat_shape = region_mat_dic["{0}__{1}".format(
                row_GenomeRegions.loc[i, "region4coolFetch"], 
                col_GenomeRegions.loc[ii, "region4coolFetch"])].shape
            

            row_GenomeRegions.loc[i, "cbins"] = cmat_shape[0]
            col_GenomeRegions.loc[ii, "cbins"] = cmat_shape[1]
            
    #print(col_GenomeRegions)
    row_GenomeRegions['chrom'] = row_GenomeRegions['chrom'].astype(str)
    col_GenomeRegions['chrom'] = col_GenomeRegions['chrom'].astype(str)
    #row_GenomeRegions['fetch_start'] = row_GenomeRegions['fetch_start'].astype(int)
    #col_GenomeRegions['fetch_start'] = col_GenomeRegions['fetch_start'].astype(int)

    r_l_regions_cMat = RegionsCmat(cmat=cMat, row_regions=row_GenomeRegions, col_regions=col_GenomeRegions)
    
    return r_l_regions_cMat
        
# import extractCisRegion as subsetCisRegion

def extractCisContact(
        clr: cooler.Cooler,
        region: str,
        extend: int = 0,
        balance: bool = False,
        divisive_weights = None,
        ) -> np.array:
    """\
    Extract cis contact matrix from the cool format Hi-C matrix.

    Parameters
    ----------
    clr: ``cooler.Cooler``
        cool format Hi-C matrix (https://github.com/open2c/cooler)
    region: ``str``
        The subset matrix row genome regions
        eg. ``"chr6:1000000-2000000"`` or ``chr6``

    extend: ``int``
        contact map extend to start and end position
    balance: ``bool``
        The ``'balance'`` parameters of ``coolMat.matrix(balance=False).fetch('chr6:119940450-123940450')``
    divisive_weights: bool, optional
        Force balancing weights to be interpreted as divisive (True) or
        multiplicative (False). Weights are always assumed to be
        multiplicative by default unless named KR, VC or SQRT_VC, in which
        case they are assumed to be divisive by default.

    Returns:
    ----------
    contact matrix: np.array
        matrix sstart with top-left
    """

    resolution = clr.binsize
    genome_region = GenomeRegion(region)

    if genome_region.chrom not in clr.chromsizes:
         logging.error(genome_region.chrom, ' is not a chrom in the cool matrix') 

    maxChromL = clr.chromsizes[genome_region.chrom]

    # ------
    if genome_region.fetch_start != None:
        genome_region.fetch_start = genome_region.fetch_start - extend*resolution
        if genome_region.fetch_start < 0:
            #genome_region.fetch_start = 0
            logging.error("Error: Extend start is less than 0, set extend=0 is ok\n")
            #sys.exit(1)

        genome_region.fetch_end = genome_region.fetch_end + extend*resolution
        if genome_region.fetch_end > maxChromL:
            #genome_region.fetch_end = maxChromL
            logging.error("Error: extend is larger than chrom length, set extend=0 is ok\n")
            #sys.exit(1)
    
    df = clr.matrix(balance=balance).fetch(genome_region.fetchRegion())
    return df




import os
import re
from logging import info, warning
from pathlib import Path
from shutil import copytree, rmtree
from tempfile import TemporaryDirectory
from typing import Any, Dict, ForwardRef, List, Union
from git import Repo


HERE = Path(__file__).parent


def _fetch_notebooks(repo_url: str) -> None:
    def copy_files(repo_path: Union[str, Path]) -> None:
        repo_path = Path(repo_path)

        for dirname in ["tutorials", "auto_examples", "gen_modules"]:
            rmtree(dirname, ignore_errors=True)  # locally re-cloning
            copytree(repo_path / "docs" / "source" / dirname, dirname)

    def fetch_remote(repo_url: str) -> None:
        info(f"Fetching notebooks from repo `{repo_url}`")
        with TemporaryDirectory() as repo_dir:
            ref = "main"
            repo = Repo.clone_from(repo_url, repo_dir, depth=1, branch=ref)
            repo.git.checkout(ref, force=True)

            copy_files(repo_dir)

    def fetch_local(repo_path: Union[str, Path]) -> None:
        info(f"Fetching notebooks from local path `{repo_path}`")
        repo_path = Path(repo_path)
        if not repo_path.is_dir():
            raise OSError(f"Path `{repo_path}` is not a directory.")

        copy_files(repo_path)

    notebooks_local_path = Path(
        os.environ.get("SQUIDPY_NOTEBOOKS_PATH", HERE.absolute().parent.parent.parent / "trackc_notebooks")
    )
    try:
        fetch_local(notebooks_local_path)
    except Exception as e:  # noqa: BLE001
        warning(f"Unable to fetch notebooks locally from `{notebooks_local_path}`, reason: `{e}`. Trying remote")
        download = int(os.environ.get("TRACKC_DOWNLOAD_NOTEBOOKS", 1))
        if not download:
            # use possibly old files, otherwise, bunch of warnings will be shown
            info(f"Not fetching notebooks from remove because `TRACKC_DOWNLOAD_NOTEBOOKS={download}`")
            return

        fetch_remote(repo_url)




from trackc.tl._getRegionsCmat import GenomeRegion
from matplotlib.axes import Axes
from typing import Union, Sequence, Optional
import pandas as pd
from trackc.pa import trackcl_11

def zoomin(ax: Optional[Axes] = None,
           row_regions: Union[pd.DataFrame, None] = None,
           zoomin_regions: Union[Sequence[str], str, None] = None,
           colors: Union[Sequence[str], None] = None,
           alpha: float = 1,
           line_on: bool = False,
           fill: bool = True,
           ):
    """\
    map_order:
        mapc mat or mat2
    """
    
    row_GRs = _region_pos(row_regions)
    zoomin_GRs = _region_pos(zoomin_regions)
    zoomin_GRs['ors'] = None
    zoomin_GRs['ore'] = None
    zoomin_GRs_cp = zoomin_GRs.copy()
    for i, row in zoomin_GRs_cp.iterrows():
        start, end = _get_zoom_origin_pos(row['chrom'], row['fetch_start'], row['fetch_end'], row_GRs)
        ss = start
        ee = end
        if row['isReverse'] == True:
            ss = end
            ee = start
        zoomin_GRs.loc[i, 'ors'] = ss
        zoomin_GRs.loc[i, 'ore'] = ee
    
    zoomin_GRs = zoomin_GRs.reset_index()

    if colors == None:
        colors = trackcl_11
    
    for i, row in zoomin_GRs.iterrows():
        x = row[['ps', 'pe', 'ore', 'ors']]
        y = [0,0,1,1]

        if fill == True:
            ax.fill(x, y, color=colors[i], alpha=alpha)

        if line_on==True:
            ax.plot((x[0], x[3]), (y[0], y[3]),color=colors[i], alpha=alpha, solid_capstyle='butt')
            ax.plot((x[1], x[2]), (y[1], y[2]),color=colors[i], alpha=alpha, solid_capstyle='butt')
        
    ax.set_axis_off()
    ax.set_xlim([0,1])
    ax.set_ylim([0,1])

def _region_pos(regions):
    GRs = pd.concat([GenomeRegion(i).GenomeRegion2df() for i in regions])
    GRs['len'] = GRs['fetch_end'] - GRs['fetch_start']
    GRs['pos_e'] = GRs['len'].cumsum()
    GRs['pos_s'] = GRs['pos_e'] - GRs['len']
    GRs_fulllen = GRs['len'].sum()
    GRs['ps'] = GRs['pos_s']/GRs_fulllen
    GRs['pe'] = GRs['pos_e']/GRs_fulllen
    return GRs


def _get_zoom_origin_pos(zoom_chrom, zooms, zoome, origin_pos_df):
    fulllen = origin_pos_df['len'].sum()
    origin_pos = origin_pos_df[origin_pos_df['chrom']==zoom_chrom]
    origin_pos = origin_pos[(zooms>=origin_pos['fetch_start']) & (zoome<=origin_pos['fetch_end'])]
    
    start = None
    end = None
    if origin_pos.shape[0] == 0:
        print('{0}:{1}-{2} not in origion regions'.format(zoom_chrom, zooms, zoome))
    else:
        ix = origin_pos.index[0]
        start_dis = origin_pos_df.loc[ix, 'pos_s'] + abs(zooms - origin_pos_df.loc[ix, 'start'])
        end_dis = origin_pos_df.loc[ix, 'pos_s'] + abs(zoome - origin_pos_df.loc[ix, 'start'])
        start = start_dis/fulllen
        end = end_dis/fulllen
    return start, end





"""Utility functions and classes
"""
import sys
import inspect
import warnings
import pandas as pd
from typing import Union, Callable, Optional, Mapping, Any, Dict, Tuple
from types import ModuleType, MethodType
from weakref import WeakSet
from functools import partial, wraps


def _getdoc(c_or_f: Union[Callable, type]) -> Optional[str]:
    if getattr(c_or_f, '__doc__', None) is None:
        return None
    doc = inspect.getdoc(c_or_f)
    if isinstance(c_or_f, type) and hasattr(c_or_f, '__init__'):
        sig = inspect.signature(c_or_f.__init__)
    else:
        sig = inspect.signature(c_or_f)

    def type_doc(name: str):
        param: inspect.Parameter = sig.parameters[name]
        cls = getattr(param.annotation, '__qualname__', repr(param.annotation))
        if param.default is not param.empty:
            return f'{cls}, optional (default: {param.default!r})'
        else:
            return cls

    return '\n'.join(
        f'{line} : {type_doc(line)}' if line.strip() in sig.parameters else line
        for line in doc.split('\n')
    )


def _one_of_ours(obj, root: str):
    return (
        hasattr(obj, "__name__")
        and not obj.__name__.split(".")[-1].startswith("_")
        and getattr(
            obj, '__module__', getattr(obj, '__qualname__', obj.__name__)
        ).startswith(root)
    )

def _descend_classes_and_funcs(mod: ModuleType, root: str, encountered=None):
    if encountered is None:
        encountered = WeakSet()
    for obj in vars(mod).values():
        if not _one_of_ours(obj, root):
            continue
        if callable(obj) and not isinstance(obj, MethodType):
            yield obj
            if isinstance(obj, type):
                for m in vars(obj).values():
                    if callable(m) and _one_of_ours(m, root):
                        yield m
        elif isinstance(obj, ModuleType) and obj not in encountered:
            encountered.add(obj)
            yield from _descend_classes_and_funcs(obj, root, encountered)


def annotate_doc_types(mod: ModuleType, root: str):
    for c_or_f in _descend_classes_and_funcs(mod, root):
        c_or_f.getdoc = partial(_getdoc, c_or_f)






# the actual API
from trackc import tl, pl, pa
from trackc.gs import (
    make_spec,
    tenon,
    savefig
)

__all__ = [
    "tl",
	"pl",
    "pa",
    "make_spec",
    "tenon"
]

import sys
sys.modules.update({f'{__name__}.{m}': globals()[m] for m in ['tl', 'pl', 'pa']})
from ._utils import annotate_doc_types
annotate_doc_types(sys.modules[__name__], 'trackc')
del sys, annotate_doc_types



import setuptools
import glob
import os
#from setuptools import setup, find_packages


def read_requirements(fname):
    with open(fname, 'r', encoding='utf-8') as file:
        return [line.rstrip() for line in file]


setuptools.setup(
    name='trackc',
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    packages=setuptools.find_packages(where='src'),
    package_dir={'': 'src'},
    py_modules=[os.path.splitext(os.path.basename(path))[0] for path in glob.glob('src/*.py')],
    include_package_data=True,
    install_requires=read_requirements('requirements.txt'),
    author="Zan Yuan",
    author_email="yfinddream@gmail.com",
    #packages=find_packages(),
    scripts=['bin/gtf2bed4trackc'],
    #include_package_data=True,
    #package_dir={'trackc': 'src/trackc'},
    #package_data={'trackc': ['qc_template.html']},
    url='http://trackc.readthedocs.io',
    description="Track view of chromosome conformation and multi-omics data",
    #long_description=open('README.rst').read(),
    long_description='x',
    #url="https://github.com/seqyuan/trackc",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from trackc.tl._getRegionsCmat import GenomeRegion
from .Virtual4C import _get_pets
from matplotlib.axes import Axes
from typing import Union, Sequence, Optional
import pandas as pd

def _mark_contact_regions_square(ax, 
                                 markRegions, 
                                 plot_region_df, 
                                 map_order, 
                                 symmetric, 
                                 linewidth, 
                                 linestyle, 
                                 linecolor, 
                                 only_cis
                                 ):
    for i, reg in enumerate(markRegions):
        reg_pos = plot_region_df[plot_region_df[reg]==1]
        x0 = reg_pos.index[0] - 0.5
        x1 = reg_pos.index[-1] + 0.5
        
        """
        ---> ln1
        |ln3 |ln2
        ---->ln4 
        
        (pos1)   _____ (pos4)
                |     |
                |     |
        (pos2)   _____ (pos3)
        """
        ln1, = ax.plot((x0, x1), (x0, x0), linestyle=linestyle, color=linecolor, linewidth=linewidth, solid_capstyle='butt')
        ln2, = ax.plot((x1, x1), (x0, x1), linestyle=linestyle, color=linecolor, linewidth=linewidth, solid_capstyle='butt')
        ln3, = ax.plot((x0, x0), (x0, x1), linestyle=linestyle, color=linecolor, linewidth=linewidth, solid_capstyle='butt')
        ln4, = ax.plot((x0, x1), (x1, x1), linestyle=linestyle, color=linecolor, linewidth=linewidth, solid_capstyle='butt')

        if symmetric == False:
            if map_order == 0:
                ln3.remove()
                ln4.remove()
            else:
                ln1.remove()
                ln2.remove()

        if only_cis==False:
            for region in markRegions[i+1:]:
                trans_reg_pos = plot_region_df[plot_region_df[region]==1]
                y0 = trans_reg_pos.index[0]
                y1 = trans_reg_pos.index[-1]
                ln5, = ax.plot((x0, x1), (y0, y0), linestyle=linestyle, color=linecolor, linewidth=linewidth, solid_capstyle='butt')
                ln6, = ax.plot((x0, x1), (y1, y1), linestyle=linestyle, color=linecolor, linewidth=linewidth, solid_capstyle='butt')
                
                ln7, = ax.plot((y0, y0), (x0, x1), linestyle=linestyle, color=linecolor, linewidth=linewidth, solid_capstyle='butt')
                ln8, = ax.plot((y1, y1), (x0, x1), linestyle=linestyle, color=linecolor, linewidth=linewidth, solid_capstyle='butt')
                
                if symmetric == False:
                    if map_order == 1:
                        ln7.remove()
                        ln8.remove()
                    if map_order == 0:
                        ln5.remove()
                        ln6.remove()
            for region in markRegions[:i]:
                trans_reg_pos = plot_region_df[plot_region_df[region]==1]
                y0 = trans_reg_pos.index[0]
                y1 = trans_reg_pos.index[-1]
                ln5, = ax.plot((y0, y0), (x0, x1), linestyle=linestyle, color=linecolor, linewidth=linewidth, solid_capstyle='butt')
                ln6, = ax.plot((y1, y1), (x0, x1), linestyle=linestyle, color=linecolor, linewidth=linewidth, solid_capstyle='butt')
                
                ln7, = ax.plot((x0, x1), (y0, y0), linestyle=linestyle, color=linecolor, linewidth=linewidth, solid_capstyle='butt')
                ln8, = ax.plot((x0, x1), (y1, y1), linestyle=linestyle, color=linecolor, linewidth=linewidth, solid_capstyle='butt')
                
                if symmetric == False:
                    if map_order == 1:
                        ln7.remove()
                        ln8.remove()
                    if map_order == 0:
                        ln5.remove()
                        ln6.remove()
    

def _mark_contact_regions_triangle(ax, 
                                   markRegions, 
                                   plot_region_df, 
                                   map_order, 
                                   symmetric, 
                                   linewidth, 
                                   linestyle, 
                                   linecolor, 
                                   only_cis
                                   ):
    for i, reg in enumerate(markRegions):
        reg_pos = plot_region_df[plot_region_df[reg]==1]

        x0 = reg_pos.index[0]
        x1 = reg_pos.index[-1]
        ln1, = ax.plot((x0, x0+(x1-x0)/2), (0, x1-x0), linestyle=linestyle, color=linecolor, linewidth=linewidth, solid_capstyle='butt')
        ln2, = ax.plot((x0+(x1-x0)/2, x1), (x1-x0, 0), linestyle=linestyle, color=linecolor, linewidth=linewidth, solid_capstyle='butt')
        
        ln3, = ax.plot((x0, x0+(x1-x0)/2), (0, (x1-x0)*-1), linestyle=linestyle, color=linecolor, linewidth=linewidth, solid_capstyle='butt')
        ln4, = ax.plot((x0+(x1-x0)/2, x1), ((x1-x0)*-1, 0), linestyle=linestyle, color=linecolor, linewidth=linewidth, solid_capstyle='butt')
        if symmetric == False:
            if map_order == 0:
                ln3.remove()
                ln4.remove()
            else:
                ln1.remove()
                ln2.remove()

        if only_cis==False:
            for region in markRegions[i+1:]:
                trans_reg_pos = plot_region_df[plot_region_df[region]==1]
                y0 = trans_reg_pos.index[0]
                y1 = trans_reg_pos.index[-1]
                pos_botom = (x1+(y0-x1)/2, y0-x1)
                pos_top = (x0+(y1-x0)/2, y1-x0)
                pos_l = (x0+(y0-x0)/2, y0-x0)
                pos_r = (x1+(y1-x1)/2, y1-x1)

                ln5, = ax.plot((pos_botom[0], pos_r[0]), (pos_botom[1], pos_r[1]), linestyle=linestyle, color=linecolor, linewidth=linewidth, solid_capstyle='butt')
                ln6, = ax.plot((pos_top[0], pos_l[0]), (pos_top[1], pos_l[1]), linestyle=linestyle, color=linecolor, linewidth=linewidth, solid_capstyle='butt')
                ln7, = ax.plot((pos_botom[0], pos_r[0]), (-pos_botom[1], -pos_r[1]), linestyle=linestyle, color=linecolor, linewidth=linewidth, solid_capstyle='butt')
                ln8, = ax.plot((pos_top[0], pos_l[0]), (-pos_top[1], -pos_l[1]), linestyle=linestyle, color=linecolor, linewidth=linewidth, solid_capstyle='butt')
                
                ln5_2, = ax.plot((pos_botom[0], pos_l[0]), (pos_botom[1], pos_l[1]), linestyle=linestyle, color=linecolor, linewidth=linewidth, solid_capstyle='butt')
                ln6_2, = ax.plot((pos_top[0], pos_r[0]), (pos_top[1], pos_r[1]), linestyle=linestyle, color=linecolor, linewidth=linewidth, solid_capstyle='butt')
                ln7_2, = ax.plot((pos_botom[0], pos_l[0]), (-pos_botom[1], -pos_l[1]), linestyle=linestyle, color=linecolor, linewidth=linewidth, solid_capstyle='butt')
                ln8_2, = ax.plot((pos_top[0], pos_r[0]), (-pos_top[1], -pos_r[1]), linestyle=linestyle, color=linecolor, linewidth=linewidth, solid_capstyle='butt')
                
                if symmetric == False:
                    if map_order == 0:
                        ln7.remove()
                        ln8.remove()
                        ln7_2.remove()
                        ln8_2.remove()
                    if map_order == 1:
                        ln5.remove()
                        ln6.remove()
                        ln5_2.remove()
                        ln6_2.remove()
                        
def _mark_contact_regions(row_regions, 
                         markRegions, 
                         binsize,
                         ax, 
                         map_type, 
                         map_order,
                         symmetric, 
                         linestyle, 
                         linecolor, 
                         linewidth,
                         only_cis):
    
    plot_region_df = _get_pets(row_regions, binsize)
 
    mark_GenomeRegions = pd.concat([GenomeRegion(i).GenomeRegion2df() for i in markRegions])
    mark_GenomeRegions_cp = mark_GenomeRegions.copy()
    for i, row in mark_GenomeRegions_cp.iterrows():
        plot_region_df[i] = 0
        bin_s = int(row['fetch_start']/binsize)
        if row['fetch_start']%binsize > 0:
            bin_s = bin_s + 1

        bin_e = int(row['fetch_end']/binsize)
        if row['fetch_end']%binsize > 0:
            bin_e = bin_e + 1
        one_region = plot_region_df[plot_region_df["chrom"]==row['chrom']]
        one_region = one_region[(one_region['chrbin']>=bin_s) & (one_region['chrbin']<=bin_e)]
        if one_region.shape[0] > 0:
            plot_region_df.loc[one_region.index, i] = 1

    if map_type == "square":
        _mark_contact_regions_square(ax, markRegions, plot_region_df, map_order, symmetric, linewidth, linestyle, linecolor, only_cis)
    elif map_type == "triangle":
        _mark_contact_regions_triangle(ax, markRegions, plot_region_df, map_order, symmetric, linewidth, linestyle, linecolor, only_cis)
    else:
        pass
       

def mapc_markline(row_regions: Union[pd.DataFrame, None] = None,
                  mark_regions: Union[Sequence[str], str, None] = None,
                  binsize: Union[int, None] = 10000,
                  ax: Optional[Axes] = None,
                  map_type: Union[str, None] = 'square',
                  map_order: int = 0,
                  symmetric: bool = False,
                  trans_ax: bool = False,
                  only_cis: bool = False,
                  linestyle: Union[str, None]='--',
                  linecolor: Union[str, None]='k',
                  linewidth: Union[int, None]=1,
                  show_regions_edge: bool = False,
              ):
    """\
    map_order:
        mapc mat or mat2
    """
    if map_order == 1:
        for child in ax.get_children():
            if isinstance(child, Axes):
                ax = child
                break
        
    if mark_regions != None:
        _mark_contact_regions(row_regions, mark_regions, binsize, ax, map_type, map_order, symmetric, linestyle, linecolor, linewidth, only_cis)

    if show_regions_edge==True:
        _plot_contact_regions_line(ax, row_regions, map_order, map_type, trans_ax, linewidth, linestyle, linecolor)

def _plot_contact_regions_line(ax, 
                               rl_regions, 
                               map_order=0,
                               map_type='triangle',
                               trans_ax=False,
                               linewidth=1,
                               linestyle='--',
                               linecolor='k'):
    if trans_ax == True:
        print('trans_ax should be False')

    xA, yA, xB, yB = 0,0,0,0
    xA2, yA2, xB2, yB2 = 0,0,0,0
    rl_regions = rl_regions.query('chrom == chrom')

    cbins = rl_regions['cbins'].to_list()
    full_len = sum(cbins)
    v_sum = 0
    for i, v in enumerate(cbins):
        v_sum = v_sum+v

        if i == len(cbins)-1:
            break

        #v2 = cbins[i+1]

        if map_type == 'triangle':
            if map_order==0:
                xA, yA, xB, yB = v_sum/2, v_sum, v_sum, 0
                xA2, yA2, xB2, yB2 = v_sum, 0, v_sum+(full_len-v_sum)/2, full_len-v_sum
            else:
                xA, yA, xB, yB = v_sum/2, -v_sum, v_sum, 0
                xA2, yA2, xB2, yB2 = v_sum, 0, v_sum+(full_len-v_sum)/2, -(full_len-v_sum)
        
        if map_type == 'square':
            if map_order==0:
                xA, yA, xB, yB = v_sum-0.5, -0.5, v_sum-0.5, v_sum-0.5
                xA2, yA2, xB2, yB2 = v_sum-0.5, v_sum-0.5, full_len-0.5, v_sum-0.5
            else:
                xA, yA, xB, yB = -0.5, v_sum-0.5, v_sum-0.5, v_sum-0.5
                xA2, yA2, xB2, yB2 = v_sum-0.5, v_sum-0.5, v_sum-0.5, full_len-0.5

        ax.plot([xA, xB], [yA, yB], linestyle, linewidth=linewidth, color=linecolor, solid_capstyle='butt')
        ax.plot([xA2, xB2], [yA2, yB2], linestyle, linewidth=linewidth, color=linecolor, solid_capstyle='butt')



